from __future__ import annotations

import dataclasses
import typing

from metadsl import *
import metadsl.typing_tools
from .rules import *

__all__ = ["Abstraction", "Variable"]

T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")


@dataclasses.dataclass(eq=False)
class Variable:
    """
    should only be equal to itself
    """

    def __str__(self):
        return hex(id(self))

    def __repr__(self):
        return f"Variable: {str(self)}"


class Abstraction(Expression, typing.Generic[T, U]):
    @expression
    def __call__(self, arg: T) -> U:
        ...

    @expression
    @classmethod
    def create(cls, var: T, body: U) -> Abstraction[T, U]:
        ...

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[T], U]) -> Abstraction[T, U]:
        v: T = cls.create_variable(Variable())
        return cls.create(v, fn(v))

    @expression
    @classmethod
    def create_variable(cls, variable: Variable) -> T:
        ...

    @expression
    def __add__(self, other: Abstraction[V, T]) -> Abstraction[V, U]:
        """
        Composes this function with another.
        (f + g)(x) == f(g(x))
        """
        ...

    @staticmethod
    @expression
    def fix(fn: Abstraction[T, T]) -> T:
        """
        Fixed pointer operator, used to define recursive functions.
        """
        return fn(Abstraction.fix(fn))

    @expression  # type: ignore
    @property
    def unfix(self) -> Abstraction[Abstraction[T, U], Abstraction[T, U]]:
        """
        Extracts the inner function created from a fixed point operator.

        If its not a fixed point funciton, returns a function that always returns the original.
        """
        ...


# Run the `from_fn` before any other rules, so that variables
# are all created before any are replaced
# This is needed if a variable is caught in an inner scope of a local function
from_fn_rule = register_pre(default_rule(Abstraction[T, U].from_fn))
# Run  the fixed point  operator  rule after all others, so that
# it is only expanded if we need it.
fix_rule = register_post(default_rule(Abstraction.fix))


def _replace(body: U, var: T, arg: T) -> U:
    """
    Replaces all instances of `var` with `arg` inside of `body`, 
    except for local bindings of `var` as declared in other `from_fn`s inside.
    """
    if body == var:
        return arg  # type: ignore
    if not isinstance(body, Expression):
        return body

    is_abstraction = (
        isinstance(body.function, metadsl.typing_tools.BoundInfer)
        and body.function.fn == Abstraction.create.fn  # type: ignore
    )
    # If is  a `from_fn` node with the same  var bound, don't try replacing its children
    if is_abstraction and body.args[0] == var:
        return body  # type: ignore
    return body.function(
        *(_replace(a, var, arg) for a in body.args),
        **{k: _replace(v, var, arg) for k, v in body.kwargs.items()},
    )


@register
@rule
def compose(vl: T, bl: U, vr: V, br: T) -> R[Abstraction[V, U]]:
    # We want to define composition for the lambda calculus
    # We start with our two functions, each with a variable and a body:
    # f = 𝜆vl.bl
    # g = 𝜆vr.br
    # We want to compute their composition:
    # 𝜆x.f(g(x))
    # We can start by replacing function application with replacing all instances of the variable in the body
    # == f(br[vr/x])
    # == bl[vl/br[vr/x]]
    # Now, what we want is to pull out the `x` replacement to the outside, so we can create another function from this
    # Assuming no overlapping variables names in the scope, we should be able to do the replacements in sequence:
    # == bl[vl/br][vr/x]
    # == 𝜆vr.bl[vl/br]
    # this checks out typing wise, which is nice!

    return (
        Abstraction[T, U].create(vl, bl)  # type: ignore
        + Abstraction[V, T].create(vr, br),
        lambda: Abstraction.create(vr, _replace(bl, vl, br)),
    )


@register  # type: ignore
@rule
def beta_reduce(var: T, body: U, arg: T) -> R[U]:
    return (Abstraction[T, U].create(var, body)(arg), lambda: _replace(body, var, arg))


@register
@rule
def unfix_normal(
    var: T, body: U
) -> R[Abstraction[Abstraction[T, U], Abstraction[T, U]]]:
    original = Abstraction.create(var, body)
    # If this is a normal abstraction, then just return the original
    return (
        original.unfix,
        Abstraction[Abstraction[T, U], Abstraction[T, U]].from_fn(lambda _: original),
    )


@register
@rule
def unfix_fixed(
    a: Abstraction[Abstraction[T, U], Abstraction[T, U]]
) -> R[Abstraction[Abstraction[T, U], Abstraction[T, U]]]:
    return (
        Abstraction.fix(a).unfix,
        a,
    )



#!/usr/bin/env python
# coding: utf-8

# In[1]:


from metadsl import *
from metadsl_core import *
from metadsl_llvm.pure import *

import metadsl_visualize

import typez
typez.SHOW_TYPES = False

# Disable fixed point expansion temporarily
run_post_rules(False)


# In[2]:


one = Integer.from_int(1)
zero = Integer.from_int(0)


# In[3]:


mod = Module.create("Main")

@FunctionThree.from_fn_recursive
def fib_more(
    fn: FunctionThree[Integer, Integer, Integer, Integer],
    n: Integer,
    a: Integer,
    b: Integer,
) -> Integer:
    pred_cont = n > one
    minus1 = n - one
    ab = a + b
    added = fn(minus1, b, ab)

    n_eq_1 = n.eq(one)
    return pred_cont.if_(added, n_eq_1.if_(b, a))

res = mod.register_function_three(fib_more)
mod, fib_more_ref = res.left(), res.right()

@FunctionOne.from_fn
def fib(n: Integer) -> Integer:
    return fib_more_ref(n, zero, one)


mod.register_function_one(fib)._ipython_display_()


# In[ ]:


fib(Integer.from_int(10))


# In[ ]:







from __future__ import annotations
import typing

from metadsl import *
from metadsl_core import *

import metadsl_visualize

T = typing.TypeVar("T")
Shape = Vec[Integer]
Indices = Vec[Integer]


class Array(Expression, typing.Generic[T]):
    @expression
    @classmethod
    def create(cls, shape: Shape, index_fn: Abstraction[Indices, T]) -> Array[T]:
        ...

    @expression
    def __matmul__(self, other: Array[T]) -> Array[T]:
        ...

    @expression
    @classmethod
    def range(cls, length: Integer) -> Array[Integer]:
        ...

    @expression
    def __add__(self, other: Array[T]) -> Array[T]:
        ...

    @expression
    def transpose(self) -> Array[T]:
        ...


@register
@rule
def _transpose(shape: Shape, idx: Abstraction[Indices, T]) -> R[Array[T]]:
    dims = shape.length
    result_shape = shape.drop(Integer.from_int(1)).append(shape[Integer.from_int(0)])

    @Abstraction.from_fn
    def result_idx(indices: Indices) -> T:
        # take last of index vector and pass to front
        last_index = indices[dims]
        rest_indices = indices.take(dims - Integer.from_int(1))

        return idx(Vec.create(last_index) + rest_indices)

    return Array.create(shape, idx).transpose(), Array.create(result_shape, result_idx)


@register
@rule
def _outer_product(
    l_shape: Shape,
    r_shape: Shape,
    l_idx: Abstraction[Indices, Integer],
    r_idx: Abstraction[Indices, Integer],
):
    l_size = l_shape.length

    @Abstraction.from_fn
    def index_fn(indices: Indices) -> Integer:
        return l_idx(indices.take(l_size)) * r_idx(indices.drop(l_size))

    return (
        Array.create(l_shape, l_idx) @ Array.create(r_shape, r_idx),
        Array.create(l_shape + r_shape, index_fn),
    )


@register
@rule
def _add_integer(
    l_shape: Shape,
    r_shape: Shape,
    l_idx: Abstraction[Indices, Integer],
    r_idx: Abstraction[Indices, Integer],
):
    @Abstraction.from_fn
    def index_fn(indices: Indices) -> Integer:
        return l_idx(indices) + r_idx(indices)

    return (
        Array.create(l_shape, l_idx) + Array.create(r_shape, r_idx),
        Array.create(l_shape, index_fn),
    )


@register
@rule
def _range_create(length: Integer):
    @Abstraction.from_fn
    def index_fn(indices: Indices) -> Integer:
        return indices[Integer.from_int(0)]

    return (Array.range(length), Array.create(Shape.create(length), index_fn))


python_string_rules = RulesRepeatFold()

rule_groups["python_string"] = python_string_rules
register_python_string = python_string_rules.append


@expression
def concat(l: str, r: str) -> str:
    ...


@register
@rule
def _concat_rule(l: str, r: str):
    return concat(l, r), lambda: l + r


@expression
def python_string_integer(s: str) -> Integer:
    ...


@expression
def integer_python_string(i: Integer) -> str:
    ...


@register
@rule
def _python_string_integer_rule(s: str):
    return integer_python_string(python_string_integer(s)), s


@expression
def python_string_shape(s: str) -> Shape:
    ...


@expression
def shape_python_string(s: Shape) -> str:
    ...


@register
@rule
def _vec_add_str(l: str, r: str):
    return (
        python_string_shape(l) + python_string_shape(r),
        lambda: python_string_shape(f"({l} + {r})"),
    )


@register
@rule
def _python_string_shape_rule(s: str):
    return shape_python_string(python_string_shape(s)), s


@expression
def python_string_array(s: str) -> Array[Integer]:
    ...


@expression
def array_python_string(a: Array[Integer]) -> str:
    ...


@register
@rule
def _python_string_array_rule(s: str):
    return array_python_string(python_string_array(s)), s


# @register_python_string
# @rule
# def _to_array_three(i: str, j: str, k: str, index_fn: Abstraction[Indices, Integer]):
#     def result():
#         return python_string_array(
#             concat(
#                 f"""
# shape = ({i}, {j}, {k})
# array = np.empty(shape)
# for i_ in range(shape[0]):
#     for j_ in range(shape[1]):
#         for k_ in range(shape[2]):
#             array[i_, j_, k_] = """,
#                 integer_python_string(
#                     index_fn(
#                         Vec.create(
#                             python_string_integer("i_"),
#                             python_string_integer("j_"),
#                             python_string_integer("k_"),
#                         )
#                     )
#                 ),
#             )
#         )

#     return (
#         Array.create(
#             Vec.create(
#                 python_string_integer(i),
#                 python_string_integer(j),
#                 python_string_integer(k),
#             ),
#             index_fn,
#         ),
#         result,
#     )


@register_python_string
@rule
def _to_vec_fn(l: str, index_fn: Abstraction[Integer, Integer]) -> R[Vec[Integer]]:
    def result() -> Vec[Integer]:
        return python_string_shape(
            concat(
                concat(
                    "[", integer_python_string(index_fn(python_string_integer("i"))),
                ),
                f"for i in range({l})]",
            )
        )

    return Vec.create_fn(python_string_integer(l), index_fn), result


@register_python_string
@rule
def _to_array(shape: str, index_fn: Abstraction[Indices, Integer]):
    def result():
        return python_string_array(
            concat(
                f"""
shape = {shape}
array = np.empty(shape)
for indices in itertools.product(*map(range, shape)):
    array[indices] = """,
                integer_python_string(index_fn(python_string_shape("indices"))),
            )
        )

    return Array.create(python_string_shape(shape), index_fn), result


@register_python_string
@rule
def _to_integer(i: int):

    return Integer.from_int(i), lambda: python_string_integer(str(i))


@register_python_string
@rule
def _add_integer_str(l: str, r: str):

    return (
        python_string_integer(l) + python_string_integer(r),
        lambda: python_string_integer(f"({l} + {r})"),
    )


@expression
def python_string_bool(s: str) -> Boolean:
    ...


@expression
def bool_python_string(b: Boolean) -> str:
    ...


@register_python_string
@rule
def _other_things_string(b: str, l: str, r: str):
    yield bool_python_string(python_string_bool(b)), b
    yield python_string_bool(b).if_(
        python_string_integer(l), python_string_integer(r)
    ), lambda: python_string_integer(f"{l} if {b} else {r}")
    yield python_string_integer(l) < python_string_integer(
        r
    ), lambda: python_string_bool(f"({l} < {r})")
    yield python_string_integer(l) - python_string_integer(
        r
    ), lambda: python_string_integer(f"({l} - {r})")


@register_python_string
@rule
def _to_shape_zero():
    return Vec[Integer].create(), lambda: python_string_shape("tuple()")


@register_python_string
@rule
def _to_shape_one(i: str):
    return Vec.create(python_string_integer(i)), lambda: python_string_shape(f"({i},)")


@register_python_string
@rule
def _to_shape_two(i: str, j: str):
    return (
        Vec.create(python_string_integer(i), python_string_integer(j)),
        lambda: python_string_shape(f"({i}, {j})"),
    )


@register_python_string
@rule
def _to_shape_three(i: str, j: str, k: str):
    return (
        Vec.create(
            python_string_integer(i), python_string_integer(j), python_string_integer(k)
        ),
        lambda: python_string_shape(f"({i}, {j}, {k})"),
    )


@register_python_string
@rule
def _getitem_str(s: str, i: str):
    return (
        python_string_shape(s)[python_string_integer(i)],
        lambda: python_string_integer(f"{s}[{i}]"),
    )


@register_python_string
@rule
def _to_shape_four(i: str, j: str, k: str, l: str):
    return (
        Vec.create(
            python_string_integer(i),
            python_string_integer(j),
            python_string_integer(k),
            python_string_integer(l),
        ),
        lambda: python_string_shape(f"({i}, {j}, {k}, {l})"),
    )


@register_python_string
@rule
def _to_shape_take_drop(s: str, i: str):
    yield (
        python_string_shape(s).take(python_string_integer(i)),
        lambda: python_string_shape(f"{s}[:{i}]"),
    )
    yield (
        python_string_shape(s).drop(python_string_integer(i)),
        lambda: python_string_shape(f"{s}[{i}:]"),
    )


@register_python_string
@rule
def _int_multiple_str(l: str, r: str):
    yield (
        python_string_integer(l) * python_string_integer(r),
        lambda: python_string_integer(f"({l} * {r})"),
    )


def ones(shape: Shape):
    @Abstraction.from_fn
    def idx_fn(indices: Indices) -> Integer:
        return Integer.from_int(1)

    return Array.create(shape, idx_fn)


python_eval_rules = RulesRepeatFold()

rule_groups["python_eval"] = python_eval_rules
register_python_eval = python_eval_rules.append


import numpy
import itertools


@expression
def array_to_ndarray(a: Array[Integer]) -> numpy.ndarray:
    ...


@register_python_eval
@rule
def _eval_array_str(s: str):
    def result():
        locals_: typing.Dict = {}
        eval(
            compile(s, filename="<string>", mode="exec"),
            {"np": numpy, "itertools": itertools},
            locals_,
        )
        return locals_["array"]

    return (
        array_to_ndarray(python_string_array(s)),
        result,
    )


# ten = Integer.from_int(10)

# res = array_to_ndarray((Array.range(ten) + ones(Vec.create(ten))) @ Array.range(ten))
# execute(res)



from __future__ import annotations

from metadsl import *
import metadsl.typing_tools
from .function import *
from .integer import *
from .abstraction import *

one = Integer.from_int(1)
zero = Integer.from_int(0)


class TestFunction:
    def test_zero_abstraction(self):
        @FunctionZero.from_fn
        def fn() -> Integer:
            return one

        assert execute(fn()) == one

    def test_one_abstraction(self):
        @FunctionOne.from_fn
        def add_one(a: Integer) -> Integer:
            return a + one

        assert (
            metadsl.typing_tools.get_type(execute(add_one.abstraction))
            == Abstraction[Integer, Integer]
        )

        assert execute(add_one.abstraction(zero)) == one

    def test_one_recursive_call(self):
        @FunctionOne.from_fn_recursive
        def factorial(fact_fn: FunctionOne[Integer, Integer], n: Integer) -> Integer:
            return n.eq(zero).if_(one, n * fact_fn(n - one))

        assert execute(factorial(zero)) == one
        assert execute(factorial(one)) == one
        assert execute(factorial(Integer.from_int(3))) == Integer.from_int(6)

    def test_two_abstraction(self):
        @FunctionTwo.from_fn
        def add(a: Integer, b: Integer) -> Integer:
            return a + b

        assert (
            metadsl.typing_tools.get_type(execute(add.abstraction))
            == Abstraction[Integer, Abstraction[Integer, Integer]]
        )

        assert execute(add.abstraction(one)(zero)) == one

    def test_two_call(self):
        @FunctionTwo.from_fn
        def add(a: Integer, b: Integer) -> Integer:
            return a + b

        assert execute(
            add(Integer.from_int(1), Integer.from_int(2))
        ) == Integer.from_int(3)

    def test_three_recursive_call(self):
        @FunctionThree.from_fn_recursive
        def fib_more(
            fn: FunctionThree[Integer, Integer, Integer, Integer],
            n: Integer,
            a: Integer,
            b: Integer,
        ) -> Integer:
            pred_cont = n > one
            minus1 = n - one
            ab = a + b
            added = fn(minus1, b, ab)

            n_eq_1 = n.eq(one)
            return pred_cont.if_(added, n_eq_1.if_(b, a))

        @FunctionOne.from_fn
        def fib(n: Integer) -> Integer:
            return fib_more(n, zero, one)

        assert execute(fib(zero)) == zero
        assert execute(fib(one)) == one
        assert execute(fib(Integer.from_int(8))) == Integer.from_int(21)



"""
This module is built on top of the abstraction

It adds the ability to have multiple arities and names attached to functions.

Because MyPy doesn't support multiple arity generics, we need a seperate class for each arity of
function.
"""
from __future__ import annotations


from metadsl import *
from .rules import *
from .abstraction import *
import typing

__all__ = ["FunctionZero", "FunctionOne", "FunctionTwo", "FunctionThree"]


T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")
X = typing.TypeVar("X")


class FunctionZero(Expression, typing.Generic[T]):
    """
    Function with zero args.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[], T]) -> FunctionZero[T]:
        return cls.create(fn.__name__, fn())

    @expression
    def __call__(self) -> T:
        ...

    @expression
    @classmethod
    def create(cls, name: str, val: T) -> FunctionZero[T]:
        ...

    @property  # type: ignore
    @expression
    def value(self) -> T:
        ...

    @expression  # type: ignore
    @property
    def name(self) -> str:
        ...


class FunctionOne(Expression, typing.Generic[T, U]):
    """
    Function with one arg.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[T], U]) -> FunctionOne[T, U]:
        return cls.create(fn.__name__, Abstraction.from_fn(fn))

    @expression
    @classmethod
    def from_fn_recursive(
        cls, fn: typing.Callable[[FunctionOne[T, U], T], U]
    ) -> FunctionOne[T, U]:
        @Abstraction.fix
        @Abstraction.from_fn
        def inner(inner_abst: Abstraction[T, U]) -> Abstraction[T, U]:
            inner_fn = cls.create(fn.__name__, inner_abst)

            @Abstraction.from_fn
            def inner(arg1: T) -> U:
                return fn(inner_fn, arg1)

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    def __call__(self, arg: T) -> U:
        ...

    @expression
    @classmethod
    def create(cls, name: str, val: Abstraction[T, U]) -> FunctionOne[T, U]:
        ...

    @expression  # type: ignore
    @property
    def abstraction(self) -> Abstraction[T, U]:
        ...

    @expression  # type: ignore
    @property
    def name(self) -> str:
        ...

    @expression  # type: ignore
    @property
    def unfix(self) -> Abstraction[FunctionOne[T, U], FunctionOne[T, U]]:
        ...


class FunctionTwo(Expression, typing.Generic[T, U, V]):
    """
    Function with two args.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[T, U], V]) -> FunctionTwo[T, U, V]:
        def inner(arg1: T) -> Abstraction[U, V]:
            def inner(arg2: U) -> V:
                return fn(arg1, arg2)

            return Abstraction.from_fn(inner)

        return cls.create(fn.__name__, Abstraction.from_fn(inner))

    @expression
    @classmethod
    def from_fn_recursive(
        cls, fn: typing.Callable[[FunctionTwo[T, U, V], T, U], V]
    ) -> FunctionTwo[T, U, V]:
        @Abstraction.fix
        @Abstraction.from_fn
        def inner(
            inner_abst: Abstraction[T, Abstraction[U, V]]
        ) -> Abstraction[T, Abstraction[U, V]]:
            inner_fn = cls.create(fn.__name__, inner_abst)

            @Abstraction.from_fn
            def inner(arg1: T) -> Abstraction[U, V]:
                @Abstraction.from_fn
                def inner(arg2: U) -> V:
                    return fn(inner_fn, arg1, arg2)

                return inner

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    def __call__(self, arg1: T, arg2: U) -> V:
        ...

    @expression
    @classmethod
    def create(
        cls, name: str, val: Abstraction[T, Abstraction[U, V]]
    ) -> FunctionTwo[T, U, V]:
        ...

    @expression  # type: ignore
    @property
    def abstraction(self) -> Abstraction[T, Abstraction[U, V]]:
        ...

    @expression  # type: ignore
    @property
    def name(self) -> str:
        ...

    @expression  # type: ignore
    @property
    def unfix(self) -> Abstraction[FunctionTwo[T, U, V], FunctionTwo[T, U, V]]:
        ...


class FunctionThree(Expression, typing.Generic[T, U, V, X]):
    """
    Function with three args.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[T, U, V], X]) -> FunctionThree[T, U, V, X]:
        @Abstraction.from_fn
        def inner(arg1: T) -> Abstraction[U, Abstraction[V, X]]:
            @Abstraction.from_fn
            def inner(arg2: U) -> Abstraction[V, X]:
                @Abstraction.from_fn
                def inner(arg3: V) -> X:
                    return fn(arg1, arg2, arg3)

                return inner

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    @classmethod
    def from_fn_recursive(
        cls, fn: typing.Callable[[FunctionThree[T, U, V, X], T, U, V], X]
    ) -> FunctionThree[T, U, V, X]:
        @Abstraction.fix
        @Abstraction.from_fn
        def inner(
            inner_abst: Abstraction[T, Abstraction[U, Abstraction[V, X]]]
        ) -> Abstraction[T, Abstraction[U, Abstraction[V, X]]]:
            inner_fn = cls.create(fn.__name__, inner_abst)

            @Abstraction.from_fn
            def inner(arg1: T) -> Abstraction[U, Abstraction[V, X]]:
                @Abstraction.from_fn
                def inner(arg2: U) -> Abstraction[V, X]:
                    @Abstraction.from_fn
                    def inner(arg3: V) -> X:
                        return fn(inner_fn, arg1, arg2, arg3)

                    return inner

                return inner

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    def __call__(self, arg1: T, arg2: U, arg3: V) -> X:
        ...

    @expression
    @classmethod
    def create(
        cls, name: str, val: Abstraction[T, Abstraction[U, Abstraction[V, X]]]
    ) -> FunctionThree[T, U, V, X]:
        ...

    @expression  # type: ignore
    @property
    def abstraction(self) -> Abstraction[T, Abstraction[U, Abstraction[V, X]]]:
        ...

    @expression  # type: ignore
    @property
    def name(self) -> str:
        ...

    @expression  # type: ignore
    @property
    def unfix(
        self,
    ) -> Abstraction[FunctionThree[T, U, V, X], FunctionThree[T, U, V, X]]:
        ...


register_pre(default_rule(FunctionZero[T].from_fn))
register_pre(default_rule(FunctionOne[T, U].from_fn))
register_pre(default_rule(FunctionTwo[T, U, V].from_fn))
register_pre(default_rule(FunctionThree[T, U, V, X].from_fn))

register_pre(default_rule(FunctionOne[T, U].from_fn_recursive))
register_pre(default_rule(FunctionTwo[T, U, V].from_fn_recursive))
register_pre(default_rule(FunctionThree[T, U, V, X].from_fn_recursive))


@register  # type: ignore
@rule
def zero_call(_: str, v: T) -> R[T]:
    return FunctionZero.create(_, v)(), v


@register  # type: ignore
@rule
def one_call(_: str, a: Abstraction[T, U], t: T) -> R[U]:
    return FunctionOne.create(_, a)(t), a(t)


@register  # type: ignore
@rule
def two_call(_: str, a: Abstraction[T, Abstraction[U, V]], t: T, u: U) -> R[V]:
    return FunctionTwo.create(_, a)(t, u), a(t)(u)


@register  # type: ignore
@rule
def three_call(
    _: str, a: Abstraction[T, Abstraction[U, Abstraction[V, X]]], t: T, u: U, v: V
) -> R[X]:
    return FunctionThree.create(_, a)(t, u, v), a(t)(u)(v)


@register  # type: ignore
@rule
def zero_value(_: str, t: T) -> R[T]:
    return FunctionZero.create(_, t).value, t


@register  # type: ignore
@rule
def one_abstraction(fn: FunctionOne[T, U]) -> R[Abstraction[T, U]]:
    return fn.abstraction, lambda: Abstraction[T, U].from_fn(lambda t: fn(t))


@register  # type: ignore
@rule
def two_abstraction(fn: FunctionTwo[T, U, V]) -> R[Abstraction[T, Abstraction[U, V]]]:
    return (
        fn.abstraction,
        lambda: Abstraction[T, Abstraction[U, V]].from_fn(
            lambda t: Abstraction[U, V].from_fn(lambda u: fn(t, u))
        ),
    )


@register  # type: ignore
@rule
def three_abstraction(
    fn: FunctionThree[T, U, V, X]
) -> R[Abstraction[T, Abstraction[U, Abstraction[V, X]]]]:
    return (
        fn.abstraction,
        lambda: Abstraction[T, Abstraction[U, Abstraction[V, X]]].from_fn(
            lambda t: Abstraction[U, Abstraction[V, X]].from_fn(
                lambda u: Abstraction[V, X].from_fn(lambda v: fn(t, u, v))
            )
        ),
    )


@register  # type: ignore
@rule
def zero_name(name: str, t: T):
    return FunctionZero.create(name, t).name, name


@register  # type: ignore
@rule
def one_name(_: str, abst: Abstraction[T, U]):
    return FunctionOne.create(_, abst).name, _


@register  # type: ignore
@rule
def two_name(_: str, abst: Abstraction[T, Abstraction[U, V]]):
    return FunctionTwo.create(_, abst).name, _


@register  # type: ignore
@rule
def three_name(_: str, abst: Abstraction[T, Abstraction[U, Abstraction[V, X]]]):
    return FunctionThree.create(_, abst).name, _


@register_pre  # type: ignore
@rule
def one_unfix(name: str, abst: Abstraction[T, U]):
    def result():
        @Abstraction.from_fn
        def inner(fn: FunctionOne[T, U]) -> FunctionOne[T, U]:
            return FunctionOne.create(name, abst.unfix(fn.abstraction))

        return inner

    return FunctionOne.create(name, abst).unfix, result


@register_pre  # type: ignore
@rule
def two_unfix(name: str, abst: Abstraction[T, Abstraction[U, V]]):
    def result():
        @Abstraction.from_fn
        def inner(fn: FunctionTwo[T, U, V]) -> FunctionTwo[T, U, V]:
            return FunctionTwo.create(name, abst.unfix(fn.abstraction))

        return inner

    return FunctionTwo.create(name, abst).unfix, result


@register_pre  # type: ignore
@rule
def three_unfix(name: str, abst: Abstraction[T, Abstraction[U, Abstraction[V, X]]]):
    def result():
        @Abstraction.from_fn
        def inner(fn: FunctionThree[T, U, V, X]) -> FunctionThree[T, U, V, X]:
            return FunctionThree.create(name, abst.unfix(fn.abstraction))

        return inner

    return FunctionThree.create(name, abst).unfix, result



import json
import html
import copy
import re
from pprint import pformat
from collections import namedtuple
from collections.abc import MutableSequence
from contextlib import contextmanager


object_entry = namedtuple("object_entry", ['objid'])
attr_entry = namedtuple("attr_entry", ["objid", "name"])
flatten_entry = namedtuple("flatten_entry", ['handle'])

object_meta = namedtuple("object_meta", ['cls', 'attrs'])


def _get_attr_entry(obj_entry, attrname):
    return attr_entry(objid=obj_entry.objid, name=attrname)


def _recreate(resmngr, value):
    """Recreate a flattened value into an object.
    """
    if isinstance(value, flatten_entry):
        meta = {
            'resmngr': resmngr,
            'handle': value.handle,
        }
        cls, attrs = resmngr[value.handle]
        value = cls(meta)
    return value


_err_migrate = (
    "Cannot migrate Managed objects between different "
    "ResourceManager instances"
)


def _make_stringify_ctx():
    return {
        'depth': 0,
        # Map object-id to a shorter id.
        'showed': {}
    }


_in_stringify = None


@contextmanager
def stringify_context(obj):
    """A context to indicate that we are nested in a `__str__` or `__repr__`.
    """
    global _in_stringify
    if not _in_stringify:
        _in_stringify = _make_stringify_ctx()
    _in_stringify['depth'] += 1
    showed = _in_stringify['showed']
    objid = obj._obj_handle
    nested = objid in showed
    if not nested:
        showed[objid] = objid.objid
    try:
        yield nested
    finally:
        _in_stringify['depth'] -= 1
        if _in_stringify['depth'] == 0:
            _in_stringify = None


def get_short_id(obj):
    objid = obj._obj_handle
    if _in_stringify:
        return _in_stringify['showed'][objid]
    else:
        return objid


class RedirectAccess(type):
    special_attrs = {'_resmngr', '_obj_handle'}

    def __new__(cls, name, bases, d):
        d['__slots__'] = cls.special_attrs
        return type.__new__(cls, name, bases, d)

    def __init__(cls, name, bases, d):
        type.__init__(cls, name, bases, d)

        def setter(self, k, v):
            if k not in cls.special_attrs:
                entry = _get_attr_entry(self._obj_handle, k)
                if isinstance(v, Managed):
                    v = self._resmngr.flatten(v)
                elif v is None:
                    pass
                elif isinstance(v, (int, float, str)):
                    pass
                elif self._resmngr.is_accepted_types(v):
                    pass
                else:
                    raise ValueError(
                        'cannot use value of type {}'.format(type(v)),
                    )

                try:
                    self._resmngr[entry] = v
                except KeyError:
                    raise AttributeError(k)
                else:
                    old = self._resmngr[self._obj_handle]
                    new = old._replace(attrs=old.attrs | {k})
                    self._resmngr[self._obj_handle] = new
            else:
                super(cls, self).__setattr__(k, v)

        def getter(self, k):
            if k not in cls.special_attrs and not k.startswith('__'):
                entry = _get_attr_entry(self._obj_handle, k)
                try:
                    v = self._resmngr[entry]
                except KeyError:
                    raise AttributeError(
                        "Missing {}.{}".format(type(self).__name__, k),
                    )
                else:
                    return _recreate(self._resmngr, v)
            else:
                return super(cls, self).__getattr__(k)

        def delter(self, k):
            if k not in cls.special_attrs and not k.startswith('__'):
                entry = _get_attr_entry(self._obj_handle, k)
                try:
                    del self._resmngr[entry]
                except KeyError:
                    raise AttributeError(k)
                else:
                    old = self._resmngr[self._obj_handle]
                    new = old._replace(attrs=old.attrs - {k})
                    self._resmngr[self._obj_handle] = new
            else:
                return super(cls, self).__delattr__(k)

        cls.__setattr__ = setter
        cls.__getattr__ = getter
        cls.__delattr__ = delter

    def __call__(cls, _meta_, *args, **kwargs):
        resmngr = _meta_['resmngr']
        obj = object.__new__(cls)
        obj._resmngr = resmngr
        if 'handle' in _meta_:
            obj._obj_handle = _meta_['handle']
        else:
            obj._obj_handle = object_entry(objid=_new_slot(resmngr))
            resmngr[obj._obj_handle] = object_meta(cls=cls, attrs=frozenset())
        return obj


def _new_slot(heap):
    try:
        return heap['__heap_slot__']
    finally:
        heap['__heap_slot__'] += 1


assume_info_is_id0 = "Info object id is assumed to be 0"


class IncompleteStateError(Exception):
        """An error indicating that the ResourceManager cannot be loaded
        from a dictionary because it is incomplete.
        """
        pass


class ResourceManager(object):
    __slots__ = ['_dct', '_infos']
    accepted_types = ()


    @classmethod
    def load_from_dict(cls, dct):
        """Make a new ResourceManager from a dictionary.
        """
        self = cls()
        self._dct = dct
        try:
            infos = self.find_object_by_id(0)
        except KeyError:
            raise IncompleteStateError("Info entry is not available.")
        assert isinstance(infos, Infos), assume_info_is_id0
        self._infos = infos
        return self

    def __init__(self):
        from tict import Tict
        # dictimpl = dict
        dictimpl = Tict
        self._dct = dictimpl()
        self._dct['__heap_slot__'] = 0
        self._infos = self.new(Infos)
        assert self.flatten(self._infos).handle.objid == 0, assume_info_is_id0

    def __getitem__(self, k):
        return self._dct[k]

    def __setitem__(self, k, v):
        self._dct[k] = v

    def __delitem__(self, k):
        del self._dct[k]

    @property
    def infos(self):
        return self._infos

    def is_accepted_types(self, instance):
        accepted = self.accepted_types
        return accepted and isinstance(instance, accepted)

    def new(self, typ_, *args, **kwargs):
        if not issubclass(typ_, Managed):
            raise TypeError("{} is not a subclass of {}".format(typ_, Managed))
        obj = typ_({'resmngr': self})
        obj.init(*args, **kwargs)
        return obj

    def new_list(self, initializer=()):
        lst = self.new(ManagedList)
        for it in initializer:
            lst.append(it)
        return lst

    def clone(self):
        return copy.deepcopy(self)

    def flatten(self, managed_obj):
        """Flatten a Managed object into a handle.
        """
        if not isinstance(managed_obj, Managed):
            raise ValueError('expecting an instance of {}'.format(Managed))
        if managed_obj._resmngr is not self:
            raise ValueError(_err_migrate)
        return flatten_entry(handle=managed_obj._obj_handle)

    def inflate(self, handle):
        """Recreate a Managed object from a handle.
        """
        if not isinstance(handle, flatten_entry):
            raise ValueError(
                'expecting an instance of {}'.format(flatten_entry),
            )
        return _recreate(self, handle)

    def dump(self):
        return pformat(self._dct)

    def iter_attributes(self):
        """Returns an iterator that yields key-value pairs which corresponds
        to attribute entries.
        """
        return iter((k, v) for k, v in self._dct.items()
                    if isinstance(k, attr_entry))

    def replace_all(self, old, new):
        """Replace all references to Managed object *old* to *new*.

        Returns
        -------
        replaced_count : int
        """
        replmap = {}
        hdl = self.flatten(new)
        oldhdl = self.flatten(old)
        for k, v in self.iter_attributes():
            if oldhdl == v:
                replmap[k] = hdl
        self._dct.update(replmap)
        return len(replmap)

    def iter_objects(self):
        """Returns an iterator that yields all the Managed objects.
        """
        for k in self._dct:
            if isinstance(k, object_entry):
                yield self.find_object_by_id(k.objid)

    def iter_referers(self, obj):
        """Returns an iterator that yields all the referer of an object.
        Precisely, it yields a 2-tuple of ``(referer_object, attrname)``
        """
        hdl = self.flatten(obj)
        for k, v in self.iter_attributes():
            if v == hdl:
                referer_id = k.objid
                referer = self.find_object_by_id(referer_id)
                attrname = k.name
                yield _iter_referers_result(referer, attrname)

    def iter_referents(self, obj):
        """Returns an iterator that yields all the referents of an object.
        Precisely, it yields a 2-tuple of ``(attrname, referent_object)``
        """
        for k in self.get_attributes(obj):
            val = self[_get_attr_entry(obj._obj_handle, k)]
            if isinstance(val, flatten_entry):
                val = self.inflate(val)
            yield _iter_referents_result(k, val)

    def get_attributes(self, obj):
        """Returns the attributes associated with a *obj*.
        """
        hdl = obj._obj_handle
        return tuple(self[hdl].attrs)

    def find_object_by_id(self, objid):
        """Get the object using given the object ID, *objid*.
        """
        return self.inflate(flatten_entry(handle=object_entry(objid)))

    def visualize_raw(self):
        """
        Returns
        -------
        (nodes, edges) : (List[Dict], List[Dict])

        """
        # Draw nodes
        nodes = []
        managed_attrs = []
        labelfmt = ('<TD PORT={!r}>{}<BR/>'
                    '<FONT POINT-SIZE="8pt">{} :: {}</FONT></TD>')
        for obj in self.iter_objects():
            src = str(obj._obj_handle)

            labelbuf = [
                '{}@{}'.format(
                    type(obj).__name__,
                    obj._obj_handle.objid
                ),
            ]

            def attr_sort(name):
                m = re.match(r"^_\d+$", name)
                if m is None:
                    return name
                else:
                    return '{:08x}'.format(int(name[1:]))

            arglist = sorted(self.get_attributes(obj), key=attr_sort)
            for attr in arglist:
                attrval = getattr(obj, attr, None)
                if attrval is None:
                    # Allow incomplete graph.
                    # Note: useful in animating revisions.
                    continue
                elif isinstance(attrval, Managed):
                    dst = "{}:this".format(attrval._obj_handle)
                    managed_attrs.append(('{}:{}'.format(src, attr), dst))
                    labelbuf.append('<TD PORT={!r}>{}</TD>'.format(attr, attr))
                else:
                    # Non-managed values
                    desc = repr(attrval)
                    label = labelfmt.format(
                        attr,
                        attr,
                        html.escape(desc),
                        type(attrval).__name__,
                    )
                    labelbuf.append(label)

            # if isinstance(obj, ManagedList) and len(labelbuf) > 8:
            #     head, tail = labelbuf[0], labelbuf[1:]
            #     tail = ''.join(map('<TR>{}</TR>'.format, tail))
            #     label = '''<
            #         <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
            #         <TR><TD PORT="this">{}</TD></TR>
            #         {}</TABLE>
            #     >'''.format(head, tail)
            # else:
            head, tail = labelbuf[0], ''.join(labelbuf[1:])
            if tail:
                tail = "<TR>{}</TR>".format(tail)
            label = '''<
                <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
                <TR><TD PORT="this" COLSPAN="{}">{}</TD></TR>
                {}</TABLE>
            >'''.format(max(1, len(labelbuf) - 1), head, tail)

            nodes.append({
                'name': src,
                'label': label,
                'shape': 'plaintext'
            })

        # Draw edges
        edges = []
        for src, dst in managed_attrs:
            edges.append({
                'tail_name': src,
                'head_name': dst,
            })

        return nodes, edges

    def visualize(self):
        """Print a DOT graph
        """
        nodes, edges = self.visualize_raw()
        return graphviz_visualize(nodes, edges)

    def internal_dict(self):
        """Returns the internal dictionary
        """
        return self._dct.copy()


def graphviz_visualize(nodes, edges):
    """Visualize nodes and edges as with graphviz

    Note: requires the `graphviz` (the python package)
    """
    import graphviz as gv

    g = gv.Digraph()
    for node in nodes:
        g.node(**node)
    for edge in edges:
        g.edge(**edge)
    return g


def graphviz_diff(last_nodes, last_edges, nodes, edges):
    """Highlight the difference between the last and current nodes and edges.
    """
    oldnodes = {n['name']: n for n in last_nodes}
    oldedges = {info['tail_name']: info['head_name'] for info in last_edges}
    for n in nodes:
        # New node
        if n['name'] not in oldnodes:
            n['color'] = 'blue'
        # Node label changed
        elif n['label'] != oldnodes[n['name']]['label']:
            n['color'] = 'orange'
    for e in edges:
        # New edge
        if e['head_name'] != oldedges.get(e['tail_name'], None):
            e['color'] = 'blue'
    return nodes, edges


def graphviz_render_revisions(resmngr, since=None, backend='d3', attrs={}):
    """Render revisions applied to **resmngr** start from **since** revision.


    Parameters
    ----------
    resmngr :
        A ResourceManager backed with Tict.
    since : tict.SavedState or None
    backend : str; defaults to "d3"
        Available options are: "d3", and "gv".
    attrs: dict; defaults to ``{}``
        For attaching additional keyword arguments to the backend.
    """
    journals = resmngr.internal_dict()

    last_nodes, last_edges = [], []
    dots = []
    for i, rev in enumerate(journals.revisions(since=since)):
        # Load dictionary as plain ResourceManager.
        # Note, it's row-polymorphic
        try:
            rev_res = ResourceManager.load_from_dict(rev)
        except IncompleteStateError:
            continue
        nodes, edges = rev_res.visualize_raw()
        diffed = graphviz_diff(last_nodes, last_edges, nodes, edges)
        g = graphviz_visualize(*diffed)
        # g.render(filename=name, format='gv', directory='anim', cleanup=True)
        dots.append(str(g))
        last_nodes, last_edges = nodes, edges

    backends = {
        'd3': _graphviz_d3_revs_backend,
        'gv': _graphviz_revs_backend,
    }
    return backends[backend](dots, **attrs)


def _graphviz_revs_backend(dots, name_prefix='resmngr_revs', dir='anim',
                           format='gv'):
    """Graphviz revision rendering backend
    """
    import graphviz as gv

    filenames = []
    places = len(str(len(dots)))
    for i, src in enumerate(dots):
        g = gv.Source(src)
        name = '{{}}_{{:0{}}}'.format(places).format(name_prefix, i)
        path = g.render(
            filename=name, format=format, directory=dir, cleanup=True,
        )
        filenames.append(path)
    return filenames


def _graphviz_d3_revs_backend(dots):
    """D3 revision rendering backend
    """
    # See https://github.com/magjac/d3-graphviz
    with open('d3dot_template.html') as fin:
        template = fin.read()
    js_dots = map(json.dumps, map(lambda x: x.splitlines(), dots))
    return template.replace('/*replaceme*/', ',\n'.join(js_dots))


_iter_referers_result = namedtuple(
    '_iter_referers_result', ['referer', 'attrname'],
)
_iter_referents_result = namedtuple(
    '_iter_referents_result', ['attrname', 'referent'],
)


class Managed(metaclass=RedirectAccess):
    """Instances of ``Managed`` must be created by ``ResourceManager.new()``.

    Subclasses should not define ``__new__`` or ``__init__``.
    Custom initialization is defined in ``init``, which return value is
    ignored.
    """
    def init(self):
        pass

    def __repr__(self):
        with stringify_context(self) as nested:
            cls = type(self).__qualname__
            if nested:
                return "@{}".format(get_short_id(self))
            else:
                attrs = self._resmngr.get_attributes(self)
                pairs = ['{}={!r}'.format(k, getattr(self, k))
                         for k in sorted(attrs)]
                return "{}@{}({})".format(
                    cls, get_short_id(self), ', '.join(pairs),
                )


class Infos(Managed):
    pass


class ManagedList(Managed):
    def init(self, values=()):
        self._count = 0
        for v in values:
            self.append(v)

    def append(self, item):
        slot = self._count
        self._count += 1
        setattr(self, '_{}'.format(slot), item)

    def extend(self, iterable):
        for i in iterable:
            self.append(i)

    def pop(self):
        item = self[-1]
        self._count -= 1
        name = '_{}'.format(self._count)
        delattr(self, name)
        return item

    def __len__(self):
        return self._count

    def __setitem__(self, idx, item):
        idx = range(self._count)[idx]
        setattr(self, '_{}'.format(idx), item)

    def __getitem__(self, idx):
        idx = range(self._count)[idx]
        return getattr(self, '_{}'.format(idx))

    def __repr__(self):
        with stringify_context(self) as nested:
            cls = type(self).__qualname__
            if nested:
                return "@{}".format(get_short_id(self))
            else:
                return "{}@{}({})".format(
                    cls, get_short_id(self), repr(list(self)),
                )


MutableSequence.register(ManagedList)


# ------------- TEST -------------------

class Test(Managed):
    def init(self):
        self.a = 1
        self.a = 2
        self.node = self._resmngr.new(Node, 123)
        self.more = self._resmngr.new_list([
            self._resmngr.new(Node, 100),
            self._resmngr.new(Node, 200),
        ])


class Node(Managed):
    def init(self, v):
        self.b = v


def main():
    s = ResourceManager()
    t = s.new(Test)
    print('t.a =', t.a)

    print(s.dump())

    print(hasattr(t, '__dict__'))

    print('t.node =', t.node)
    t.node.b = 321
    print('t.node.b =', t.node.b)

    # print(pickle.loads(pickle.dumps(s)))
    # print(pickle.loads(pickle.dumps(t)))

    print(s.dump())

    print(t.more[0].b)
    print(t.more[1].b)


if __name__ == '__main__':
    main()




from __future__ import annotations


from metadsl import *
from .boolean import *


class TestBoolean:
    def test_if(self):
        assert execute(Boolean.create(True).if_(0, 1)) == 0
        assert execute(Boolean.create(False).if_(0, 1)) == 1



"""
Useful for to convert object to boxed types.
"""
from __future__ import annotations
import typing

from metadsl import *
from .maybe import *
from .rules import *

__all__ = ["Converter", "convert_identity_rule", "convert_to_maybe"]
T = typing.TypeVar("T")


class Converter(Expression, typing.Generic[T]):
    @expression
    @classmethod
    def convert(cls, value: object) -> Maybe[T]:
        """
        Converts a value to a certain type.
        """
        ...


@register_convert
@rule
def convert_identity_rule(value: T) -> R[Maybe[T]]:
    """
    When the value is an instance of the type being converted, we can convert it.
    """
    yield Converter[T].convert(value), Maybe.just(value)
    yield Converter[T].convert(Maybe.just(value)), Maybe.just(value)
    yield Converter[T].convert(Maybe[T].nothing()), Maybe[T].nothing()


@register_convert
@rule
def convert_to_maybe(x: object) -> R[Maybe[Maybe[T]]]:
    return (
        Converter[Maybe[T]].convert(x),
        lambda: Maybe.just(
            Maybe[T].nothing() if x is None else Converter[T].convert(x)
        ),
    )



from __future__ import annotations

from metadsl import *
from .conversion import *
from .rules import *
from .maybe import *
from .boolean import *

__all__ = ["Integer"]


class Integer(Expression):
    @expression
    @classmethod
    def from_int(cls, i: int) -> Integer:
        ...

    @expression
    def __add__(self, other: Integer) -> Integer:
        ...

    @expression
    def __sub__(self, other: Integer) -> Integer:
        ...

    @expression
    def __mul__(self, other: Integer) -> Integer:
        ...

    @expression
    def eq(self, other: Integer) -> Boolean:
        ...

    @expression
    def __lt__(self, other: Integer) -> Boolean:
        ...

    @expression
    def __le__(self, other: Integer) -> Boolean:
        ...

    @expression
    def __gt__(self, other: Integer) -> Boolean:
        ...

    @expression
    def __ge__(self, other: Integer) -> Boolean:
        ...


@register
@rule
def integer_math(l: int, r: int) -> R[Integer]:
    yield Integer.from_int(l) + Integer.from_int(r), lambda: Integer.from_int(l + r)
    yield Integer.from_int(l) - Integer.from_int(r), lambda: Integer.from_int(l - r)
    yield Integer.from_int(l) * Integer.from_int(r), lambda: Integer.from_int(l * r)


@register
@rule
def integer_comparison(l: int, r: int) -> R[Boolean]:
    li = Integer.from_int(l)
    ri = Integer.from_int(r)
    yield li.eq(ri), lambda: Boolean.create(l == r)
    yield li < (ri), lambda: Boolean.create(l < r)
    yield li <= (ri), lambda: Boolean.create(l <= r)
    yield li > (ri), lambda: Boolean.create(l > r)
    yield li >= (ri), lambda: Boolean.create(l >= r)


@register_convert
@rule
def convert_integer(i: int) -> R[Maybe[Integer]]:
    return Converter[Integer].convert(i), lambda: Maybe.just(Integer.from_int(i))



from __future__ import annotations

import typing

from metadsl import *
from .integer import *
from .maybe import *
from .conversion import *
from .abstraction import *
from .pair import *
from .rules import *

T = typing.TypeVar("T")
U = typing.TypeVar("U")

__all__ = ["Vec"]


class Vec(Expression, typing.Generic[T]):
    """
    A tuple of homogonous types.
    """

    @expression
    def __getitem__(self, index: Integer) -> T:
        ...

    @expression
    @classmethod
    def create(cls, *items: T) -> Vec[T]:
        ...

    @expression
    def map(self, fn: Abstraction[T, U]) -> Vec[U]:
        ...

    @expression
    def append(self, x: T) -> Vec[T]:
        ...

    @expression
    def fold(self, initial: U, fn: Abstraction[U, Abstraction[T, U]]) -> U:
        """
        Like this:

            def fold(xs, initial, fn):
                ret = initial
                for x in xs:
                    ret = fn(ret)(x)
                return ret
        """

    @expression
    def __add__(self, other: Vec[T]) -> Vec[T]:
        ...

    @expression  # type: ignore
    @property
    def length(self) -> Integer:
        ...

    @expression
    def take(self, i: Integer) -> Vec[T]:
        ...

    @expression
    def drop(self, i: Integer) -> Vec[T]:
        ...

    @expression
    @classmethod
    def create_fn(cls, length: Integer, fn: Abstraction[Integer, T]) -> Vec[T]:
        ...


@register
@rule
def create_fn_rules(
    l: Integer,
    fn: Abstraction[Integer, T],
    x: T,
    l1: Integer,
    fn1: Abstraction[Integer, T],
    n: Integer,
):
    v = Vec.create_fn(l, fn)
    v1 = Vec.create_fn(l1, fn1)
    one = Integer.from_int(1)
    yield v[l], fn(l)
    yield v.append(x), Vec.create_fn(
        l + one, Abstraction[Integer, T].from_fn(lambda i: (i < l).if_(fn(i), x))
    )
    yield v + v1, Vec.create_fn(
        l + l1,
        Abstraction[Integer, T].from_fn(lambda i: (i < l).if_(fn(i), fn1(i - l1))),
    )
    yield v.length, l
    yield v.take(n), Vec.create_fn(l - n, fn)
    yield v.drop(n), Vec.create_fn(
        l - n, Abstraction[Integer, T].from_fn(lambda i: fn(i + n)),
    )


@register  # type: ignore
@rule
def length_rule(xs: typing.Sequence[T]) -> R[Integer]:
    return (Vec[T].create(*xs).length, lambda: Integer.from_int(len(xs)))


@register  # type: ignore
@rule
def take_drop_rule(xs: typing.Sequence[T], i: int) -> R[Vec[T]]:
    yield (
        Vec[T].create(*xs).take(Integer.from_int(i)),
        lambda: Vec[T].create(*xs[:i]),
    )
    yield (
        Vec[T].create(*xs).drop(Integer.from_int(i)),
        lambda: Vec[T].create(*xs[i:]),
    )


@register  # type: ignore
@rule
def add_rule(ls: typing.Sequence[T], rs: typing.Sequence[T]) -> R[Vec[T]]:
    return (Vec[T].create(*ls) + Vec[T].create(*rs), Vec[T].create(*ls, *rs))


@register  # type: ignore
@rule
def getitem(i: int, xs: typing.Sequence[T]) -> R[T]:
    return (Vec[T].create(*xs)[Integer.from_int(i)], lambda: xs[i])


@register
@rule
def map(fn: Abstraction[T, U], xs: typing.Sequence[T]) -> R[Vec[U]]:
    return (
        Vec[T].create(*xs).map(fn),  # type: ignore
        lambda: (Vec.create(*(fn(x) for x in xs))) if xs else Vec[U].create(),
    )


@register
@rule
def append(xs: typing.Sequence[T], x: T) -> R[Vec[T]]:
    return (Vec[T].create(*xs).append(x), lambda: Vec.create(*xs, x))


# def tuple_all(t: Vec[Maybe[T]]) -> Maybe[Vec[T]]:
#     # accumulate a maybe tuple of T, representng the partial list so far
#     @Abstraction.from_fn
#     def fn(tpl: Maybe[Vec[T]]) -> Abstraction[Maybe[T], Maybe[Vec[T]]]:
#         @Abstraction.from_fn
#         def inner(x: Maybe[T]) -> Maybe[Vec[T]]:
#             @Abstraction.from_fn
#             def tpl_matched(just_tpl: Vec[T]) -> Maybe[Vec[T]]:

#                 @Abstraction.from_fn
#                 def x_matched(just_x: T) -> Maybe[Vec[T]]:
#                     return just_tpl.append(just_x)

#                 return x.match(
#                     Maybe[Vec[T]].nothing(),
#                     x_matched
#                 )
#             return tpl.match(
#                 tpl,
#                 tpl_matched
#             )
#         return inner

#     return t.fold(Maybe.just(Vec[T].create()), fn)


@register_convert
@rule
def convert_vec(xs: object) -> R[Maybe[Vec[T]]]:
    def replacement() -> Maybe[Vec[T]]:
        res = Maybe.just(Vec[T].create())
        # Only convert tuples
        if not isinstance(xs, tuple):
            raise NoMatch
        for x in xs:  # type: ignore

            @Abstraction.from_fn
            def fn(b: Pair[Vec[T], T]) -> Vec[T]:
                return b.left.append(b.right)

            res = (res & Converter[T].convert(x)).map(fn)  # type: ignore
        return res

    return Converter[Vec[T]].convert(xs), replacement




from __future__ import annotations
import typing
import ctypes


from metadsl import *
from metadsl_core import *
from .ir import *
from .ctypes import *
from .llvmlite_binding import *
from .ir_llvmlite import *

__all__ = ["llvm_integration_rules", "compile_function", "compile_functions"]


llvm_integration_rules = RulesRepeatFold()
register_integration = llvm_integration_rules.append


@expression
def concat_strings(l: str, r: str) -> str:
    return l + r


register(default_rule(concat_strings))


def make_c_wrapper(mod_ref: ModRef, original_fn_ref: FnRef) -> typing.Tuple[str, Fn]:
    """
    Creates a new function that wraps the old function,
    making sure the calling convention is default.
    """
    new_name = concat_strings("entry_", original_fn_ref.name)
    fn_ref = mod_ref.fn(new_name, original_fn_ref.type,)
    block_ref = fn_ref.block(True, "entry")
    return (
        new_name,
        fn_ref.fn(block_ref.ret(block_ref.call(original_fn_ref, fn_ref.arguments))),
    )


@expression
def llvm_to_c_fn_type(fn_tp: FnType) -> CFunctionType:
    ...


@expression
def llvm_to_c_type(tp: Type) -> CType:
    ...


@register_integration
@rule
def llvm_to_c_fn_type_rule(ret_tp: Type, arg_tps: typing.Sequence[Type]):
    return (
        llvm_to_c_fn_type(FnType.create(ret_tp, *arg_tps)),
        CFunctionType.create(llvm_to_c_type(ret_tp), *(map(llvm_to_c_type, arg_tps))),
    )


@register_integration
@rule
def llvm_to_c_fn_int():
    return (llvm_to_c_type(Type.create_int(32)), CType.c_int())


@expression
def compile_function(
    mod: Mod, fn_ref: FnRef, optimization: int = 1,
) -> typing.Callable:
    new_name, wrapper_fn = make_c_wrapper(mod.ref, fn_ref)
    module_ref = ModuleRef.create(mod_str(mod.add_fn(wrapper_fn)))
    module_ref = module_ref.optimize(optimization)
    engine = ExecutionEngine.create(module_ref)
    return llvm_to_c_fn_type(fn_ref.type)(engine.get_function_address(new_name))


def compile_functions(
    mod_ref: ModRef, fn: Fn, *fns: Fn, optimization: int = 1,
) -> typing.Callable:
    mod = mod_ref.mod(fn, *fns)
    return compile_function(mod, fn.ref)


register_integration(default_rule(compile_function))



from __future__ import annotations


from metadsl import *
from .abstraction import *
from .maybe import *


class Int(Expression):
    @expression
    def __mul__(self, other: Int) -> Int:
        ...

    @expression
    @classmethod
    def from_int(cls, i: int) -> Int:
        ...


class TestMaybe:
    def test_matches(self):
        @Abstraction.from_fn
        def double_int(i: Int) -> Int:
            return Int.from_int(2) * i

        assert execute(
            Maybe.just(Int.from_int(10)).match(Int.from_int(5), double_int)
        ) == Int.from_int(2) * Int.from_int(10)
        assert execute(
            Maybe[Int].nothing().match(Int.from_int(5), double_int)
        ) == Int.from_int(5)

    def test_map(self):
        @Abstraction.from_fn
        def double_int(i: Int) -> Int:
            return Int.from_int(2) * i

        assert execute(Maybe.just(Int.from_int(10)).map(double_int)) == Maybe[Int].just(
            Int.from_int(2) * Int.from_int(10)
        )
        assert execute(Maybe[Int].nothing().map(double_int)) == Maybe[Int].nothing()



import IPython.core.display
import IPython
import typing

from metadsl import *
from .typez import *

__all__ = ["execute_and_visualize"]


T = typing.TypeVar("T")


def execute_and_visualize(ref: ExpressionReference, rule: Rule) -> object:
    """
    Returns the replaced version of this expression and also displays the execution trace.
    """
    expression_display = ExpressionDisplay(ref)

    # Only display expressions if in notebook, not in shell
    if (
        IPython.get_ipython()
        and IPython.get_ipython().__class__.__name__ == "ZMQInteractiveShell"
    ):
        IPython.core.display.display(expression_display)

    # Update the typez display as we execute the rules
    for replacement in rule(ref):
        expression_display.update(replacement.rule, replacement.label)
    return ref.expression


def monkeypatch():
    """
    Monkeypatches Expression should it displays the result as well as each intermediate step.
    """
    Expression._ipython_display_ = _expression_ipython_display  # type: ignore
    # only change if we are in a kernel
    if IPython.get_ipython():
        execute.execute = execute_and_visualize  # type: ignore


def _expression_ipython_display(self):
    res = execute(self)
    # Only display result if we get back a non expression object
    if not isinstance(res, Expression):
        IPython.core.display.display(res)


monkeypatch()



from __future__ import annotations


from metadsl import *
from .conversion import *
from .either import *
from .abstraction import *
from .rules import all_rules
from .maybe import *


class Int(Expression):
    @expression
    def __mul__(self, other: Int) -> Int:
        ...

    @expression
    @classmethod
    def from_str(cls, str: Str) -> Int:
        ...

    @expression
    @classmethod
    def from_int(cls, i: int) -> Int:
        ...


class Str(Expression):
    @expression
    @classmethod
    def from_str(cls, s: str) -> Str:
        ...


IntStr = Either[Int, Str]


class TestMatches:
    def test_matches(self):
        @Abstraction.from_fn
        def double_int(i: Int) -> Int:
            return Int.from_int(2) * i

        str_to_int = Abstraction.from_fn(Int.from_str)

        assert execute(
            IntStr.left(Int.from_int(10)).match(double_int, str_to_int)
        ) == Int.from_int(2) * Int.from_int(10)

        assert execute(
            IntStr.right(Str.from_str("10")).match(double_int, str_to_int)
        ) == Int.from_str(Str.from_str("10"))


@rule
def convert_to_int(i: int) -> R[Maybe[Int]]:
    return Converter[Int].convert(i), lambda: Maybe.just(Int.from_int(i))


@rule
def convert_to_str(s: str) -> R[Maybe[Str]]:
    return Converter[Str].convert(s), lambda: Maybe.just(Str.from_str(s))


convert_rules = RulesRepeatFold(convert_to_str, convert_to_int, all_rules)


class TestConvert:
    def test_convert_int(self):
        assert execute(Converter[Int].convert(123), convert_rules) == Maybe.just(
            Int.from_int(123)
        )

    def test_convert_str(self):
        assert execute(Converter[Str].convert("hi"), convert_rules) == Maybe.just(
            Str.from_str("hi")
        )

    def test_convert_either_int(self):
        assert execute(Converter[IntStr].convert(123), convert_rules) == Maybe.just(
            IntStr.left(Int.from_int(123))
        )

    def test_convert_either_str(self):
        assert execute(Converter[IntStr].convert("hi"), convert_rules) == Maybe.just(
            IntStr.right(Str.from_str("hi"))
        )



from __future__ import annotations
import typing

from metadsl import *
from .conversion import *
from .maybe import *
from .abstraction import *
from .rules import *

__all__ = ["Either"]

T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")


class Either(Expression, typing.Generic[T, U]):
    @expression
    @classmethod
    def left(cls, value: T) -> Either[T, U]:
        ...

    @expression
    @classmethod
    def right(cls, value: U) -> Either[T, U]:
        ...

    @expression
    def match(self, left: Abstraction[T, V], right: Abstraction[U, V]) -> V:
        ...


@register  # type: ignore
@rule
def either_match(l: Abstraction[T, V], r: Abstraction[U, V], t: T, u: U) -> R[V]:
    yield Either[T, U].left(t).match(l, r), lambda: l(t)  # type: ignore
    yield Either[T, U].right(u).match(l, r), lambda: r(u)  # type: ignore


@register_convert
@rule
def convert_to_either(x: object) -> R[Maybe[Either[T, U]]]:
    """
    Converting to an either should try converting to both types.
    If either matches, then that should be the result. If neither
    can be converted, then the result should be nothing.
    """

    convert_left = (
        Converter[T]  # type: ignore
        .convert(x)
        .map(Abstraction.from_fn(Either[T, U].left))
    )
    convert_right = Converter[U].convert(x).map(Abstraction.from_fn(Either[T, U].right))

    return (Converter[Either[T, U]].convert(x), convert_left | convert_right)



import metadsl as m
import metadsl_core as mc
import metadsl_llvm as ml


def test_fib():
    ##
    # Constants
    ##

    int_type = ml.Type.create_int(32)
    zero = ml.ValueExpr.from_value(ml.Value.constant(int_type, 0))
    one = ml.ValueExpr.from_value(ml.Value.constant(int_type, 1))

    ##
    # Module
    ##
    mod_ref = ml.ModRef.create("fib")

    @ml.llvm_fn(mod_ref, ml.FnType.create(int_type, int_type, int_type, int_type))
    @mc.FunctionThree.from_fn_recursive
    def fib_more(
        self: mc.FunctionThree[ml.ValueExpr, ml.ValueExpr, ml.ValueExpr, ml.ValueExpr],
        n: ml.ValueExpr,
        a: ml.ValueExpr,
        b: ml.ValueExpr,
    ) -> ml.ValueExpr:
        return (n > one).if_(self(n - one, b, a + b), (n.eq(one).if_(b, a)))

    @ml.llvm_fn(mod_ref, ml.FnType.create(int_type, int_type))
    @mc.FunctionOne.from_fn
    def fib(n: ml.ValueExpr) -> ml.ValueExpr:
        return fib_more(n, zero, one)

    metadsl_fn = m.execute(
        ml.compile_functions(mod_ref, ml.to_llvm(fib), ml.to_llvm(fib_more))
    )
    assert metadsl_fn(10) == 55


def test_add():
    int_type = ml.Type.create_int(32)
    one = ml.ValueExpr.from_value(ml.Value.constant(int_type, 1))

    mod_ref = ml.ModRef.create("add")

    @ml.llvm_fn(mod_ref, ml.FnType.create(int_type, int_type, int_type))
    @mc.FunctionTwo.from_fn
    def add(l: ml.ValueExpr, r: ml.ValueExpr) -> ml.ValueExpr:
        return l + r + one

    real_fn = m.execute(ml.compile_functions(mod_ref, ml.to_llvm(add)))
    assert real_fn(10, 11) == 22



from __future__ import annotations
import metadsl


class Number(metadsl.Expression):
    @metadsl.expression
    def __add__(self, other: Number) -> Number:
        ...

    @metadsl.expression
    @classmethod
    def from_int(cls, i: int) -> Number:
        ...


@metadsl.rule
def add_zero(y: Number):
    yield Number.from_int(0) + y, y
    yield y + Number.from_int(0), y


assert metadsl.execute(
    Number.from_int(0) + Number.from_int(10), add_zero
) == Number.from_int(10)



"""
Generate LLVM IR
"""

# TODO: What if repr of value changes over time like for IR?

from __future__ import annotations
import typing

import llvmlite.ir as ir

from metadsl import *
from metadsl_core import *
from .ir import *

__all__ = ["ir_llvmlite_rules", "mod_str"]

ir_llvmlite_rules = RulesRepeatFold()
register_llvmlite = ir_llvmlite_rules.append


@hash_value.register
def hash_llvmlite_type(value: ir.Type) -> int:
    """
    Add custom hash for LLVM types, because otherwise they all hash to the same.
    """
    return hash((type(value), value._to_string()))


@hash_value.register
def hash_llvmlite_module(mod: ir.Module) -> int:
    """
    Add custom hash to also take into account contents.
    """
    return hash((mod, str(mod)))


@hash_value.register
def hash_llvmlite_function(fn: ir.Function) -> int:
    return hash((fn, str(fn)))


@hash_value.register
def hash_llvmlite_builder(builder: ir.IRBuilder) -> int:
    return hash((builder, str(builder)))


def union_types(type: ir.Type, *types: ir.Type) -> ir.Type:
    for tp in types:
        if tp != type:
            raise NotImplementedError(f"Cannot merge types {tp} and {type}")
    return type


# Overwrite ir builder str representation
ir.IRBuilder.__str__ = lambda self: f"IRBuilder: {hex(id(self))}\n{self.block}"


@expression
def mod_str(mod: Mod) -> str:
    ...


@expression
def box_mod(mod: ir.Module) -> Mod:
    ...


@expression
def box_fn(fn: ir.Function) -> Fn:
    ...


@expression
def box_terminate(builder: ir.IRBuilder) -> Terminate:
    ...


@expression
def box_value(value: ir.Value) -> Value:
    ...


@expression
def box_block_ref(builder: ir.IRBuilder) -> BlockRef:
    ...


@expression
def box_fn_ref(fn: ir.Function) -> FnRef:
    ...


@expression
def box_mod_ref(fn: ir.Module) -> ModRef:
    ...


@expression
def box_fn_type(fn_tp: ir.FunctionType) -> FnType:
    ...


@expression
def box_type(tp: ir.Type) -> Type:
    ...


@register_llvmlite
@rule
def type_create_int_box(bits: int):
    return Type.create_int(bits), lambda: box_type(ir.IntType(bits))


@register_llvmlite
@rule
def function_type_box_1(ret: ir.Type, arg1: ir.Type) -> R[FnType]:
    return (
        FnType.create(box_type(ret), box_type(arg1)),
        lambda: box_fn_type(ir.FunctionType(ret, (arg1,))),
    )


@register_llvmlite
@rule
def function_type_box_2(ret: ir.Type, arg1: ir.Type, arg2: ir.Type) -> R[FnType]:
    return (
        FnType.create(box_type(ret), box_type(arg1), box_type(arg2)),
        lambda: box_fn_type(ir.FunctionType(ret, (arg1, arg2))),
    )


@register_llvmlite
@rule
def function_type_box_3(
    ret: ir.Type, arg1: ir.Type, arg2: ir.Type, arg3: ir.Type
) -> R[FnType]:
    return (
        FnType.create(box_type(ret), box_type(arg1), box_type(arg2), box_type(arg3)),
        lambda: box_fn_type(ir.FunctionType(ret, (arg1, arg2, arg3))),
    )


@register_llvmlite
@rule
def box_mod_ref_create(name: str):
    return ModRef.create(name), lambda: box_mod_ref(ir.Module(name))


@register_llvmlite
@rule
def box_fn_ref_create(
    mod: ir.Module, fn_tp: ir.FunctionType, name: str, calling_convention: str
):
    def inner() -> FnRef:
        fn = ir.Function(mod, fn_tp, name)
        fn.calling_convention = calling_convention
        return box_fn_ref(fn)

    return (
        box_mod_ref(mod).fn(name, box_fn_type(fn_tp), calling_convention),
        inner,
    )


@register_llvmlite
@rule
def fn_ref_arguments(fn: ir.Function):
    return (
        box_fn_ref(fn).arguments,
        lambda: Vec.create(*(box_value(a) for a in fn.args)),
    )


@register_llvmlite
@rule
def block_ref_create(
    is_first: bool, fn: ir.Function, name: typing.Union[str, None], _: _Uniq,
):
    def inner() -> BlockRef:
        # If this is not the first block, make sure the first has already been appeneded
        if not is_first and not fn.blocks:
            raise NoMatch()
        return box_block_ref(ir.IRBuilder(fn.append_basic_block(name or "")))

    return (box_fn_ref(fn)._block(is_first, name, _), inner)


@register_llvmlite
@rule
def block_ref_ret(builder: ir.IRBuilder, value: ir.Value):
    def inner() -> Terminate:
        builder.ret(value)
        return box_terminate(builder)

    return box_block_ref(builder).ret(box_value(value)), inner


@register_llvmlite
@rule
def block_ref_icmp_signed(
    builder: ir.IRBuilder, operator: str, l: ir.Value, r: ir.Value
):
    return (
        box_block_ref(builder).icmp_signed(operator, box_value(l), box_value(r)),
        lambda: box_value(builder.icmp_signed(operator, l, r)),
    )


@register_llvmlite
@rule
def block_ref_call_1(builder: ir.IRBuilder, function: ir.Function, arg1: ir.Value):
    return (
        box_block_ref(builder).call(box_fn_ref(function), Vec.create(box_value(arg1))),
        lambda: box_value(builder.call(function, (arg1,))),
    )


@register_llvmlite
@rule
def block_ref_call_2(
    builder: ir.IRBuilder, function: ir.Function, arg1: ir.Value, arg2: ir.Value
):
    return (
        box_block_ref(builder).call(
            box_fn_ref(function), Vec.create(box_value(arg1), box_value(arg2)),
        ),
        lambda: box_value(builder.call(function, (arg1, arg2))),
    )


@register_llvmlite
@rule
def block_ref_call_3(
    builder: ir.IRBuilder,
    function: ir.Function,
    arg1: ir.Value,
    arg2: ir.Value,
    arg3: ir.Value,
):
    return (
        box_block_ref(builder).call(
            box_fn_ref(function),
            Vec.create(box_value(arg1), box_value(arg2), box_value(arg3)),
        ),
        lambda: box_value(builder.call(function, (arg1, arg2, arg3))),
    )


@register_llvmlite
@rule
def builder_cbranch(
    builder: ir.IRBuilder, cond: ir.Value, true: ir.IRBuilder, false: ir.IRBuilder
):
    def inner() -> Terminate:
        builder.cbranch(cond, true.block, false.block)
        return box_terminate(builder)

    return (
        box_block_ref(builder).cbranch(
            box_value(cond), box_block_ref(true), box_block_ref(false)
        ),
        inner,
    )


@register_llvmlite
@rule
def builder_branch(builder: ir.IRBuilder, target: ir.IRBuilder):
    def inner() -> Terminate:
        builder.branch(target)
        return box_terminate(builder)

    return (
        box_block_ref(builder).branch(box_block_ref(target)),
        inner,
    )


@register_llvmlite
@rule
def builder_add(builder: ir.IRBuilder, left: ir.Value, right: ir.Value):
    builder_ = box_block_ref(builder)
    return (
        builder_.add(box_value(left), box_value(right)),
        lambda: box_value(builder.add(left, right)),
    )


@register_llvmlite
@rule
def builder_sub(builder: ir.IRBuilder, left: ir.Value, right: ir.Value):
    builder_ = box_block_ref(builder)
    return (
        builder_.sub(box_value(left), box_value(right)),
        lambda: box_value(builder.sub(left, right)),
    )


@register_llvmlite
@rule
def builder_phi_two(
    builder: ir.IRBuilder,
    builder_1: ir.IRBuilder,
    value_1: ir.Value,
    builder_2: ir.IRBuilder,
    value_2: ir.Value,
) -> R[Value]:
    def inner() -> Value:
        phi = builder.phi(union_types(value_1.type, value_2.type))
        phi.add_incoming(value_1, builder_1.block)
        phi.add_incoming(value_2, builder_2.block)
        return box_value(phi)

    return (
        box_block_ref(builder).phi(
            Pair.create(box_value(value_1), box_block_ref(builder_1)),
            Pair.create(box_value(value_2), box_block_ref(builder_2)),
        ),
        inner,
    )


@register_llvmlite
@rule
def value_constant(tp: ir.Type, value: typing.Any) -> R[Value]:
    return (
        Value.constant(box_type(tp), value),
        lambda: box_value(ir.Constant(tp, value)),
    )


@register_llvmlite
@rule
def create_box_fn(
    fn: ir.Function,
    builder: ir.IRBuilder,
    terminate: Terminate,
    terminates: typing.Sequence[Terminate],
):
    fn_ref = box_fn_ref(fn)
    # Remove blocks that have already been added
    yield fn_ref.fn(box_terminate(builder), *terminates), fn_ref.fn(*terminates)

    # If we have added all, used fn_ref
    yield fn_ref.fn(), box_fn(fn)


@register_llvmlite
@rule
def create_box_mod(
    mod: ir.Module, ir_fn: ir.Function, fn: Fn, fns: typing.Sequence[Fn]
):
    mod_ref = box_mod_ref(mod)
    yield mod_ref.mod(box_fn(ir_fn), *fns), mod_ref.mod(*fns)
    yield mod_ref.mod(), box_mod(mod)


@register_llvmlite
@rule
def mod_str_rule(mod: ir.Module):
    return mod_str(box_mod(mod)), lambda: str(mod)



"""
We have a set of core rules that we want to execute all the time.

We also have a number of named rules, for conversions.
"""

import collections
import typing

from metadsl import *

__all__ = [
    "register",
    "register_convert",
    "register_unbox",
    "register_numpy_engine",
    "register_post",
    "register_pre",
    "register_optimize",
    "run_post_rules",
    "rule_groups",
    "all_rules",
]


core_rules = RulesRepeatFold()
register = core_rules.append


# Rules that have to take place before any other rules
core_pre_rules = RulesRepeatFold()
register_pre = core_pre_rules.append

# Rules that have to take place only after all other rules
core_post_rules = RulesSequenceFold()
register_post = core_post_rules.append


# Conversions from untyped values to typed values
convert_rules = RulesRepeatFold()
register_convert = convert_rules.append


unbox_rules = RulesRepeatFold()
register_unbox = unbox_rules.append


numpy_engine = RulesRepeatFold()
register_numpy_engine = numpy_engine.append

optimize_rules = RulesRepeatFold()
register_optimize = optimize_rules.append


rule_groups: typing.DefaultDict[str, Rule] = collections.defaultdict(
    core=core_rules,
    convert=convert_rules,
    optimize=optimize_rules,
    unbox=unbox_rules,
    execute=numpy_engine,
)

RUN_POST_RULES = True


def run_post_rules(should_run: bool) -> None:
    global RUN_POST_RULES
    RUN_POST_RULES = should_run


def all_rules(expression: ExpressionReference) -> typing.Iterable[Replacement]:
    sub_rules: typing.List[Rule] = [core_pre_rules]
    rules_in_order: typing.List[Rule] = []
    for k, v in rule_groups.items():
        sub_rules.append(v)
        rules_in_order.append(
            CollapseReplacementsRule(k, RulesRepeatSequence(*sub_rules))
        )
    ordered = RuleInOrder(*rules_in_order)
    if RUN_POST_RULES:
        return RulesRepeatSequence(ordered, core_post_rules)(expression)
    return ordered(expression)


execute.default_rule = all_rules  # type: ignore




"""
Dataflow based LLVMLite IR to create LLVMLite IR.
"""

from __future__ import annotations

import typing
import functools

from metadsl import *
from metadsl_core import *

from .ir import *
from .integration import *

__all__ = ["ir_context_rules", "llvm_fn", "to_llvm", "ValueExpr"]

ir_context_rules = RulesRepeatFold()
register_context = ir_context_rules.append


BlockCtx = Pair[BlockRef, Fn]
ValueCtx = Pair[Value, BlockCtx]


class ValueExpr(Expression):
    @expression
    def build(self, block_ctx: BlockCtx) -> ValueCtx:
        """
        Given some block, add the values to this block by adding instructions to it, or making
        new blocks on this function.
        """

        ...

    @expression
    def __add__(self, right: ValueExpr) -> ValueExpr:
        ...

    @expression
    def __sub__(self, right: ValueExpr) -> ValueExpr:
        ...

    @expression
    def if_(self, true: ValueExpr, false: ValueExpr) -> ValueExpr:
        ...

    @expression
    @classmethod
    def from_value(cls, val: Value) -> ValueExpr:
        ...

    def __lt__(self, other: ValueExpr) -> ValueExpr:
        return self.icmp_signed("<", other)

    def __gt__(self, other: ValueExpr) -> ValueExpr:
        return self.icmp_signed(">", other)

    def eq(self, other: ValueExpr) -> ValueExpr:
        return self.icmp_signed("==", other)

    @expression
    def icmp_signed(self, operator: str, right: ValueExpr) -> ValueExpr:
        ...


AllFns = typing.Union[
    FunctionOne[ValueExpr, ValueExpr],
    FunctionTwo[ValueExpr, ValueExpr, ValueExpr],
    FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr],
]

SomeFn = typing.TypeVar("SomeFn", bound=AllFns)


@expression
def from_llvm_fn_1(fn: Fn) -> FunctionOne[ValueExpr, ValueExpr]:
    ...


@expression
def from_llvm_fn_2(fn: Fn) -> FunctionTwo[ValueExpr, ValueExpr, ValueExpr]:
    ...


@expression
def from_llvm_fn_3(fn: Fn) -> FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr]:
    ...


def llvm_fn(mod_ref: ModRef, fn_tp: FnType) -> typing.Callable[[SomeFn], SomeFn]:
    def inner(fn: SomeFn, mod_ref=mod_ref, fn_tp=fn_tp) -> SomeFn:
        if isinstance(fn, FunctionOne):
            return llvm_fn_1(fn, mod_ref, fn_tp)  # type: ignore
        elif isinstance(fn, FunctionTwo):  # type: ignore
            return llvm_fn_2(fn, mod_ref, fn_tp)  # type: ignore
        return llvm_fn_3(fn, mod_ref, fn_tp)  # type: ignore

    return inner


@expression
def llvm_fn_1(
    fn: FunctionOne[ValueExpr, ValueExpr], mod_ref: ModRef, fn_tp: FnType
) -> FunctionOne[ValueExpr, ValueExpr]:
    ...


@expression
def llvm_fn_2(
    fn: FunctionTwo[ValueExpr, ValueExpr, ValueExpr], mod_ref: ModRef, fn_tp: FnType
) -> FunctionTwo[ValueExpr, ValueExpr, ValueExpr]:
    ...


@expression
def llvm_fn_3(
    fn: FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr],
    mod_ref: ModRef,
    fn_tp: FnType,
) -> FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr]:
    ...


def to_llvm(fn: AllFns) -> Fn:
    if isinstance(fn, FunctionOne):
        return to_llvm_fn_1(fn)
    elif isinstance(fn, FunctionTwo):
        return to_llvm_fn_2(fn)
    return to_llvm_fn_3(fn)


@typing.overload
def _fn_n_rule(
    mod_ref: ModRef, fn_n: FunctionOne[ValueExpr, ValueExpr], fn_tp: FnType,
) -> R[FunctionOne[ValueExpr, ValueExpr]]:
    ...


@typing.overload
def _fn_n_rule(
    mod_ref: ModRef, fn_n: FunctionTwo[ValueExpr, ValueExpr, ValueExpr], fn_tp: FnType,
) -> R[FunctionTwo[ValueExpr, ValueExpr, ValueExpr]]:
    ...


@typing.overload
def _fn_n_rule(
    mod_ref: ModRef,
    fn_n: FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr],
    fn_tp: FnType,
) -> R[FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr]]:
    ...


def _fn_n_rule(mod_ref: ModRef, fn_n: AllFns, fn_tp: FnType,) -> R[AllFns]:
    fn_ref = mod_ref.fn(fn_n.name, fn_tp, "fastcc")

    # Use this to pass into fixed point operator so it can call itself
    tmp_fn = fn_ref.fn()
    fn_expr: AllFns
    if isinstance(fn_n, FunctionOne):
        fn_expr = llvm_fn_1(fn_n, mod_ref, fn_tp)
        fn_res = fn_n.unfix(from_llvm_fn_1(tmp_fn))(
            ValueExpr.from_value(fn_ref.arguments[Integer.from_int(0)])
        )
    elif isinstance(fn_n, FunctionTwo):
        fn_expr = llvm_fn_2(fn_n, mod_ref, fn_tp)
        fn_res = fn_n.unfix(from_llvm_fn_2(tmp_fn))(
            ValueExpr.from_value(fn_ref.arguments[Integer.from_int(0)]),
            ValueExpr.from_value(fn_ref.arguments[Integer.from_int(1)]),
        )
    else:
        fn_expr = llvm_fn_3(fn_n, mod_ref, fn_tp)
        fn_res = fn_n.unfix(from_llvm_fn_3(tmp_fn))(
            ValueExpr.from_value(fn_ref.arguments[Integer.from_int(0)]),
            ValueExpr.from_value(fn_ref.arguments[Integer.from_int(1)]),
            ValueExpr.from_value(fn_ref.arguments[Integer.from_int(2)]),
        )
    res_val, res_block_ctx = fn_res.build(
        Pair.create(fn_ref.block(True), fn_ref.fn())
    ).spread
    block_ref, fn = res_block_ctx.spread
    fn = fn.add_block(block_ref.ret(res_val))
    if isinstance(fn_n, FunctionOne):
        fn_n = from_llvm_fn_1(fn)
    elif isinstance(fn_n, FunctionTwo):
        fn_n = from_llvm_fn_2(fn)
    else:
        fn_n = from_llvm_fn_3(fn)
    return (
        fn_expr,
        fn_n,
    )


@register_context
@rule
def llvm_fn_to_from_llvm_fn_1(
    fn_n: FunctionOne[ValueExpr, ValueExpr], mod_ref: ModRef, fn_tp: FnType
) -> R[FunctionOne[ValueExpr, ValueExpr]]:
    return _fn_n_rule(mod_ref, fn_n, fn_tp)


@register_context
@rule
def llvm_fn_to_from_llvm_fn_2(
    fn_n: FunctionTwo[ValueExpr, ValueExpr, ValueExpr], mod_ref: ModRef, fn_tp: FnType
) -> R[FunctionTwo[ValueExpr, ValueExpr, ValueExpr]]:
    return _fn_n_rule(mod_ref, fn_n, fn_tp)


@register_context
@rule
def llvm_fn_to_from_llvm_fn_3(
    fn_n: FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr],
    mod_ref: ModRef,
    fn_tp: FnType,
) -> R[FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr]]:
    return _fn_n_rule(mod_ref, fn_n, fn_tp)


@expression
def to_llvm_fn_1(fn: FunctionOne[ValueExpr, ValueExpr]) -> Fn:
    ...


@expression
def to_llvm_fn_2(fn: FunctionTwo[ValueExpr, ValueExpr, ValueExpr]) -> Fn:
    ...


@expression
def to_llvm_fn_3(fn: FunctionThree[ValueExpr, ValueExpr, ValueExpr, ValueExpr]) -> Fn:
    ...


@register_context
@rule
def to_from_llvm_fn_1(fn: Fn) -> R[Fn]:
    yield to_llvm_fn_1(from_llvm_fn_1(fn)), fn
    yield to_llvm_fn_2(from_llvm_fn_2(fn)), fn
    yield to_llvm_fn_3(from_llvm_fn_3(fn)), fn


@register_context
@rule
def build_call_1(fn: Fn, arg: ValueExpr, block_ctx: BlockCtx) -> R[ValueCtx]:
    target = from_llvm_fn_1(fn)(arg).build(block_ctx)
    arg_v, block_ctx = arg.build(block_ctx).spread
    res = block_ctx.left.call(fn.ref, Vec.create(arg_v))
    return (target, Pair.create(res, block_ctx))


@register_context
@rule
def build_call_2(
    fn: Fn, arg1: ValueExpr, arg2: ValueExpr, block_ctx: BlockCtx
) -> R[ValueCtx]:
    target = from_llvm_fn_2(fn)(arg1, arg2).build(block_ctx)
    arg1_v, block_ctx = arg1.build(block_ctx).spread
    arg2_v, block_ctx = arg2.build(block_ctx).spread
    res = block_ctx.left.call(fn.ref, Vec.create(arg1_v, arg2_v))
    return (target, Pair.create(res, block_ctx))


@register_context
@rule
def build_call_3(
    fn: Fn, arg1: ValueExpr, arg2: ValueExpr, arg3: ValueExpr, block_ctx: BlockCtx
) -> R[ValueCtx]:
    target = from_llvm_fn_3(fn)(arg1, arg2, arg3).build(block_ctx)
    arg1_v, block_ctx = arg1.build(block_ctx).spread
    arg2_v, block_ctx = arg2.build(block_ctx).spread
    arg3_v, block_ctx = arg3.build(block_ctx).spread
    res = block_ctx.left.call(fn.ref, Vec.create(arg1_v, arg2_v, arg3_v))
    return (target, Pair.create(res, block_ctx))


@register_context
@rule
def build_add(l_exp: ValueExpr, r_exp: ValueExpr, block_ctx: BlockCtx) -> R[ValueCtx]:
    target = (l_exp + r_exp).build(block_ctx)

    l, block_ctx = l_exp.build(block_ctx).spread
    r, block_ctx = r_exp.build(block_ctx).spread
    res = block_ctx.left.add(l, r)
    return (target, Pair.create(res, block_ctx))


@register_context
@rule
def build_sub(l_exp: ValueExpr, r_exp: ValueExpr, block_ctx: BlockCtx) -> R[ValueCtx]:
    target = (l_exp - r_exp).build(block_ctx)

    l, block_ctx = l_exp.build(block_ctx).spread
    r, block_ctx = r_exp.build(block_ctx).spread
    res = block_ctx.left.sub(l, r)
    return (target, Pair.create(res, block_ctx))


@register_context
@rule
def build_icmp_signed(
    l_exp: ValueExpr, r_exp: ValueExpr, operator: str, block_ctx: BlockCtx,
) -> R[ValueCtx]:
    target = l_exp.icmp_signed(operator, r_exp).build(block_ctx)
    l, block_ctx = l_exp.build(block_ctx).spread
    r, block_ctx = r_exp.build(block_ctx).spread
    res = block_ctx.left.icmp_signed(operator, l, r)
    return (target, Pair.create(res, block_ctx))


def switch_block(
    block_ctx: BlockCtx, finished_block: Terminate, new_block: BlockRef
) -> BlockCtx:
    return Pair.create(new_block, block_ctx.right.add_block(finished_block))


@register_context
@rule
def build_cbranch(
    cond_exp: ValueExpr, true_exp: ValueExpr, false_exp: ValueExpr, block_ctx: BlockCtx,
) -> R[ValueCtx]:
    """
    Builds a cbranch from two expressions, by creating a block for the true and false
    expressions. In each block, the true and false are evaluated and then .
    Then, there is a after block that will load the right operation using phi.

    Based on https://llvm.org/docs/tutorial/OCamlLangImpl5.html#if-then-else
    """
    target = cond_exp.if_(true_exp, false_exp).build(block_ctx)

    # First build conditional expression on curent block, to get conditional value
    cond_res, block_ctx = cond_exp.build(block_ctx).spread

    # Then add new blocks
    fn_ref = block_ctx.right.ref
    true_block_ref = fn_ref.block(False, "cbranch_true")
    false_block_ref = fn_ref.block(False, "cbranch_false")
    after_block_ref = fn_ref.block(False, "cbranch_after")

    # Then cbranch to true/false blocks based on result
    initial_terminate = block_ctx.left.cbranch(
        cond_res, true_block_ref, false_block_ref
    )

    # Now switch to the true block to create that one
    block_ctx = switch_block(block_ctx, initial_terminate, true_block_ref)
    true_res, block_ctx = true_exp.build(block_ctx).spread
    true_terminate = block_ctx.left.branch(after_block_ref)

    # Finish this block and go to false block
    block_ctx = switch_block(block_ctx, true_terminate, false_block_ref)
    false_res, block_ctx = false_exp.build(block_ctx).spread
    false_terminate = block_ctx.left.branch(after_block_ref)

    # Now switch to the last after block
    block_ctx = switch_block(block_ctx, false_terminate, after_block_ref)
    final_value = block_ctx.left.phi(
        Pair.create(false_res, false_block_ref), Pair.create(true_res, true_block_ref)
    )

    return (target, Pair.create(final_value, block_ctx))


@register_context
@rule
def build_constant(value: Value, block_ctx: BlockCtx) -> R[ValueCtx]:
    return (ValueExpr.from_value(value).build(block_ctx), Pair.create(value, block_ctx))




"""
Replacement rules are allow you to define transformation of expressions.

This file provides a number of functions that allows you to combine rules and apply them
to the whole tree.

This allows you to take a rule that optionally replaces one node and map it throughout the tree
and keep calling it until it no longer matches.
"""

from __future__ import annotations

import typing
import dataclasses
from .expressions import *
from .normalized import *

__all__ = [
    "Rule",
    "execute",
    "RulesRepeatSequence",
    "RulesRepeatFold",
    "RulesSequenceFold",
    "RuleSequence",
    "RuleInOrder",
    "RuleFold",
    "RuleRepeat",
    "Replacement",
    "CollapseReplacementsRule",
]
T = typing.TypeVar("T")


@dataclasses.dataclass
class Replacement:
    """
    Replacement is a record of a replacement that happened.
    """

    # The name of the rule that was executed
    rule: str
    # An optional label for this replacement
    label: typing.Optional[str] = None


# takes in an expression object and returns a number of replacements executed on the expression
Rule = typing.Callable[[ExpressionReference], typing.Iterable[Replacement]]


@dataclasses.dataclass
class Executor:
    # Function that takes a rule and an expression and executes it
    execute: typing.Callable[[ExpressionReference, Rule], object]
    # rule that is used if none is passed in
    default_rule: Rule

    def __call__(self, expr: T, rule: typing.Optional[Rule] = None) -> T:
        execute: typing.Callable[  # type: ignore
            [ExpressionReference, Rule], object
        ] = self.execute  # type: ignore
        default_rule: Rule = self.default_rule  # type: ignore
        rule = rule or default_rule
        assert rule
        return typing.cast(
            T,
            execute(
                ExpressionReference.from_expression(ExpressionFolder()(expr)), rule
            ),
        )


def _execute_all(ref: ExpressionReference, rule: Rule) -> object:
    """
    Call a replacement rule many times, returning the last result whole.
    """
    for replacement in rule(ref):
        pass
    return ref.expression


# Execute should be called on an expression to get the result
# External modules can set a custom `execute` function or `default_rule`
# to modify the behavior
execute = Executor(_execute_all, lambda e: [])


T_Rule = typing.TypeVar("T_Rule", bound=Rule)


@dataclasses.dataclass
class CollapseReplacementsRule:
    """
    Takes in an existing rule and adds a final state with a custom label.
    """

    label: str
    rule: Rule

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        replacement = None
        for replacement in self.rule(expr):  # type: ignore
            yield replacement
        if replacement:
            yield Replacement(rule="", label=self.label)


@dataclasses.dataclass(init=False)
class RulesRepeatSequence:
    """
    This takes in a list of rules and repeatedly applies them until no more match.
    """

    rules: typing.Tuple[Rule, ...]
    _rule: Rule

    def __init__(self, *rules: Rule):
        self.rules = rules
        self._update_rule()

    def _update_rule(self):
        execute_all = RuleSequence(self.rules)
        execute_many_times = RuleRepeat(execute_all)
        self._rule = execute_many_times  # type: ignore

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        rule: Rule = self._rule  # type: ignore
        return rule(expr)

    def append(self, rule: T_Rule) -> T_Rule:
        self.rules += (rule,)
        self._update_rule()
        return rule


@dataclasses.dataclass(init=False)
class RulesSequenceFold:

    rules: typing.Tuple[Rule, ...]
    _rule: Rule

    def __init__(self, *rules: Rule):
        self.rules = rules
        self._update_rule()

    def _update_rule(self):
        execute_all = RuleSequence(self.rules)
        execute_fold = RuleFold(execute_all)
        self._rule = execute_fold  # type: ignore

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        rule: Rule = self._rule  # type: ignore
        return rule(expr)

    def append(self, rule: T_Rule) -> T_Rule:
        self.rules += (rule,)
        self._update_rule()
        return rule


@dataclasses.dataclass(init=False)
class RulesRepeatFold:
    """
    This takes in a list of rules and repeatedly applies them recursively to the
    tree, until no more match anywhere.
    """

    rules: typing.Tuple[Rule, ...]
    _rule: Rule

    def __init__(self, *rules: Rule):
        self.rules = rules
        self._update_rule()

    def _update_rule(self):
        execute_all = RuleSequence(self.rules)
        execute_fold = RuleFold(execute_all)
        execute_fold_many_times = RuleRepeat(execute_fold)
        self._rule = execute_fold_many_times  # type: ignore

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        rule: Rule = self._rule  # type: ignore
        return rule(expr)

    def append(self, rule: T_Rule) -> T_Rule:
        self.rules += (rule,)
        self._update_rule()
        return rule


@dataclasses.dataclass(init=False)
class RuleInOrder:
    """
    Returns a new replacement rule that executes each of the rules in order,
    returning a new expression if any of them replaced.
    """

    rules: typing.Tuple[Rule, ...]

    def __init__(self, *rules: Rule):
        self.rules = rules

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        for rule in self.rules:
            for replacement in rule(expr):
                yield replacement


@dataclasses.dataclass
class RuleSequence:
    """
    Returns a new replacement rule that tries each of the replacement rules in sequence, returning the result of the first that matches.
    """

    rules: typing.Tuple[Rule, ...]

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        for rule in self.rules:
            replacement = None
            for replacement in rule(expr):
                yield replacement
            if replacement:
                return


@dataclasses.dataclass
class RuleFold:
    """
    Returns the first replacement found by starting at the top of the expression tree
    and then recursing down into its leaves.
    """

    rule: Rule

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        rule: Rule = self.rule  # type: ignore
        for child_ref in expr.descendents:
            for replacement in rule(child_ref):
                yield replacement
                return


@dataclasses.dataclass
class RuleRepeat:
    """
    Returns a new replacement rule that repeatedly calls the replacement rule
    """

    _rule: Rule
    max_calls: int = 1000

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Replacement]:
        i = 0
        rule: Rule = self._rule  # type: ignore

        for i in range(self.max_calls):
            replaced = False
            for replacement in rule(expr):
                replaced = True
                yield replacement
            if not replaced:
                return
        raise RuntimeError(
            f"Exceeded maximum number of repitions, rule: {rule}, expr: {expr}"  # type: ignore
        )



"""
Normalized expressions, for deduping and single replacements.
"""
from __future__ import annotations
import dataclasses
import typing
import itertools
import collections
import functools
import IPython.core.display
import igraph

from .expressions import *


__all__ = [
    "ExpressionReference",
    "Children",
    "Hash",
    "hash_value",
]

Hash = typing.NewType("Hash", str)


class Graph(igraph.Graph):
    """
    Graph of all expression



    Vertex attributes:
    * `name`: string of hash of expression or None if not computed yet
    * `expression`: expression object

    Edge Attributes:
    * `index`: int or string
    """

    def __init__(self, expr: object):
        super().__init__(directed=True)

        self.fully_add_expression(expr, None)
        self.assert_integrity()

    def _repr_svg_(self):
        return self.plot_custom()._repr_svg_()

    def plot_custom(self):
        return igraph.plot(
            self,
            layout=self.layout_sugiyama(),
            vertex_label=[
                f'{v.index}: {v["expression"].function if isinstance(v["expression"], Expression) else v["expression"]}'
                for v in self.vs
            ],
            edge_label=self.es["index"],
            # vertex_shape="hidden",
        )

    def fully_add_expression(
        self,
        expr: object,
        replace: typing.Optional[typing.Tuple[Hash, object]],
        *parent_ids: int,
    ) -> Hash:
        # should never be a child of one of its parents, or else we have a cycle
        assert id(expr) not in parent_ids
        children = frozenset(
            (
                index,
                self.fully_add_expression(
                    child_expression, replace, id(expr), *parent_ids
                ),
            )
            for index, child_expression in expression_children(expr)
        )
        hash_ = Hash(
            str(
                hash(
                    (
                        expr.function
                        if isinstance(expr, Expression)
                        else hash_value(expr),
                        children,
                    )
                )
            )
        )
        # If we are replacing a hash with a new expression and this is the hash of current expression, use the new one instead
        # This means we have added a bunch of nodes we dont need possibly, so we can delete those at the end
        if replace and hash_ == replace[0]:
            return self.fully_add_expression(replace[1], None, *parent_ids)
        try:
            self.lookup(hash_)
        except ValueError:
            v = self.add_vertex(expression=expr, name=hash_)

            for index, child_hash in children:
                assert isinstance(expr, Expression)
                child_v = self.lookup(child_hash)
                self.add_edge(v, child_v, index=index)
                if isinstance(index, int):
                    expr.args[index] = child_v["expression"]
                else:
                    expr.kwargs[index] = child_v["expression"]
        return Hash(hash_)

    def lookup(self, hash_: Hash) -> igraph.Vertex:
        return self.vs.find(name=hash_)

    def replace_root(self, expr: object):
        self.delete_vertices(self.vs)
        self.fully_add_expression(expr, None)
        self.assert_integrity()

    def replace_child(self, expr: object, prev_index: int) -> None:
        # Compute previous hash on the fly instead of passing it in b/c even though previous index
        # must stay constant, the previous hash could have changed between we created this sub-graph
        # and after the rule when we are replacing the expression
        root_expression = self.root_vertex["expression"]

        prev_expr = self.vs[prev_index]["expression"]
        # clear graph
        self.delete_vertices(self.vs)
        prev_hash = self.fully_add_expression(prev_expr, None)

        self.delete_vertices(self.vs)
        new_hash = self.fully_add_expression(root_expression, (prev_hash, expr))
        # Remove all vertice not children of new root node
        self.delete_vertices(
            set(self.vs.indices) - set(self.subcomponent(new_hash, igraph.OUT))
        )
        self.assert_integrity()

    def assert_integrity(self):
        assert self.is_dag()
        # Verify that this is one connected graph (not multiple roots)
        assert self.is_connected(igraph.WEAK)

        # Assert hashes and ids are unique
        hashes = self.vs["name"]
        assert len(hashes) == len(set(hashes))
        # assert edges are unique, with respect to source
        edges = [(e.source, e["index"]) for e in self.es]
        assert len(edges) == len(set(edges))

        for v in self.vs:
            expr = v["expression"]
            for e in v.out_edges():
                idx = e["index"]
                child_expr = (
                    expr.args[idx] if isinstance(idx, int) else expr.kwargs[idx]
                )
                assert id(child_expr) == id(e.target_vertex["expression"])

    @property
    def root_vertex(self) -> igraph.Vertex:
        return self.vs.select(_indegree_eq=0)[0]


@dataclasses.dataclass
class ExpressionReference:
    """
    This is a data structure that holds onto a graph of expression
    and has a pointer to one node, to represent it.
    """

    _graph: Graph
    # Top level vertex if the top level is not root, else None
    _optional_index: typing.Optional[int]

    @classmethod
    def from_expression(cls, expr: object) -> ExpressionReference:
        """
        Create a new reference from an expression
        """
        graph = Graph(expr)
        # top level expression will be first added vertex
        return cls(graph, None)

    @property
    def _is_root(self):
        return self._optional_index is None

    def replace(self, new_expression: object) -> None:
        """
        Replace this expression with a new one
        """

        if self._is_root:
            self._graph.replace_root(new_expression)
        else:
            self._graph.replace_child(
                new_expression, typing.cast(int, self._optional_index)
            )
            # Reset index after replacing b/c we don't know index of newly replaced child
            # (we could, but not sure we need it since we only replace once then through this object away)
            self._optional_index = None

    @property
    def _index(self) -> int:
        return self._graph.root_vertex.index if self._is_root else self._optional_index

    @property
    def _vertex(self) -> igraph.Vertex:
        return self._graph.vs[self._index]

    @property
    def hash(self) -> Hash:
        """
        Returns the Hash of the top level expression
        """
        return Hash(self._vertex["name"])

    @property
    def expression(self) -> object:
        """
        Returns the expression this references
        """
        return self._vertex["expression"]

    @property
    def children(self) -> Children:
        """
        Returns the direct children of this node if it has any.
        """
        args = {}
        kwargs = {}
        for e in self._graph.es[self._graph.incident(self.hash, igraph.OUT)]:
            index = e["index"]
            target_hash = e.target_vertex["name"]
            if isinstance(index, int):
                args[index] = target_hash
            else:
                kwargs[index] = target_hash
        return Children(
            kwargs=kwargs,
            args=tuple(
                id_ for index, id_ in sorted(args.items(), key=lambda kv: kv[0])
            ),
        )

    @property
    def descendents(self) -> typing.Iterable[ExpressionReference]:
        """
        Returns all nodes of this expression reference. If any of them
        get their expression replaced, it will replace that sub-node of this expression.

        If this is the root node, will return them in topo order, otherwise will be arbitrary
        """
        # If we are the root node return all topological with root nodes last
        if self._is_root:
            indices = self._graph.topological_sorting(igraph.IN)
        # Otherwise return all subcomponents
        else:
            indices = self._graph.subcomponent(self._optional_index, igraph.OUT)
        return [ExpressionReference(self._graph, i) for i in indices]


@dataclasses.dataclass(frozen=True)
class Children:
    args: typing.Tuple[Hash, ...]
    kwargs: typing.Dict[str, Hash]

    def _references(
        self,
    ) -> typing.Iterable[typing.Tuple[typing.Union[str, int], Hash]]:
        return itertools.chain(enumerate(self.args), self.kwargs.items())


@functools.singledispatch
def hash_value(value: object) -> int:
    """
    Computes some hash for a value that should be stable. Either use the built in hash, or if we cannot
    (like the object is mutable) then use the id.

    It's a single dispatch function so that you can register custom hashes for objects you don't control.
    """
    try:
        return hash((type(value), value))
    except TypeError:
        return hash((type(value), id(value)))


def expression_children(
    expr: object,
) -> typing.Iterable[typing.Tuple[typing.Union[int, str], object]]:
    if not isinstance(expr, Expression):
        return
    yield from enumerate(expr.args)
    yield from expr.kwargs.items()



from __future__ import annotations
import typing

from metadsl import *
from .rules import *

__all__ = ["Pair"]

T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")


class Pair(Expression, typing.Generic[T, U]):
    @expression  # type: ignore
    @property
    def left(self) -> T:
        ...

    @expression  # type: ignore
    @property
    def right(self) -> U:
        ...

    @expression
    def set_left(self, left: T) -> Pair[T, U]:
        ...

    @expression
    def set_right(self, right: U) -> Pair[T, U]:
        ...

    @expression
    @classmethod
    def create(cls, left: T, right: U) -> Pair[T, U]:
        ...

    @property
    def spread(self) -> typing.Tuple[T, U]:
        return self.left, self.right


@register  # type: ignore
@rule
def pair_left(l: T, r: U) -> R[T]:
    """
    >>> execute(Pair.create(10, 20).left)
    10
    """
    return Pair.create(l, r).left, l


@register  # type: ignore
@rule
def pair_right(l: T, r: U) -> R[U]:
    """
    >>> execute(Pair.create(10, 20).right)
    20
    """
    return Pair.create(l, r).right, r


@register  # type: ignore
@rule
def pair_set(l: T, r: U, new_l: T, new_r: U) -> R[Pair[T, U]]:
    """
    >>> execute(Pair.create(10, 20).set_left(5).left)
    5
    >>> execute(Pair.create(10, 20).set_left(5).right)
    20
    >>> execute(Pair.create(10, 20).set_right(5).left)
    10
    >>> execute(Pair.create(10, 20).set_right(5).right)
    5
    """
    yield Pair.create(l, r).set_left(new_l), Pair.create(new_l, r)
    yield Pair.create(l, r).set_right(new_r), Pair.create(l, new_r)




from __future__ import annotations


from metadsl import *
from .conversion import *
from .either import *
from .abstraction import *
from .rules import *
from .maybe import *
from .vec import *
from .integer import *


class Int(Expression):
    @expression
    @classmethod
    def from_int(cls, i: int) -> Int:
        ...


class Str(Expression):
    @expression
    @classmethod
    def from_str(cls, s: str) -> Str:
        ...


@rule
def convert_to_int(i: int, s: str) -> R[Maybe[Int]]:
    yield Converter[Int].convert(i), lambda: Maybe.just(Int.from_int(i))
    yield Converter[Int].convert(s), Maybe[Int].nothing()


@rule
def convert_to_str(i: int, s: str) -> R[Maybe[Str]]:
    yield Converter[Str].convert(s), lambda: Maybe.just(Str.from_str(s))
    yield Converter[Str].convert(i), Maybe[Str].nothing()


convert_rules = RulesRepeatFold(  # type: ignore
    convert_to_str, convert_to_int, execute.default_rule  # type: ignore
)

e = lambda e: execute(e, convert_rules)


class TestVec:
    def test_getitem(self):
        assert execute(Vec.create(10, 11)[Integer.from_int(1)]) == 11

    def test_append(self):
        assert execute(Vec.create(10).append(11)) == Vec.create(10, 11)

    def test_convert_empty(self):
        assert e(Converter[Vec[Int]].convert(())) == Maybe.just(Vec[Int].create())

    def test_convert_items(self):
        # TODO: Figure out why I need to manually annotate the maybe
        assert e(Converter[Vec[Int]].convert((1, 2))) == Maybe[Vec[Int]].just(
            Vec.create(Int.from_int(1), Int.from_int(2))
        )

    def test_invalid_conversion(self):
        assert e(Converter[Vec[Int]].convert((1, "hi"))) == Maybe[Vec[Int]].nothing()
        assert e(Converter[Vec[Int]].convert(("hi", 1))) == Maybe[Vec[Int]].nothing()
        assert e(Converter[Vec[Int]].convert(("hi",))) == Maybe[Vec[Int]].nothing()



# from __future__ import annotations

# import typing

# from .wraps import *
# from .expressions import *


# T_some = typing.TypeVar("T_some", bound="SomeExpr")


# class SomeExpr(Expression):
#     @classmethod
#     @expression
#     def create(cls: typing.Type[T_some]) -> T_some:
#         ...


# @expression
# def create_expr() -> Expression:
#     ...


# @expression
# def fn(a: SomeExpr, b: Expression) -> SomeExpr:
#     ...


# class SomeWrap(SomeExpr):
#     pass


# @wrap(fn)
# def fn_wrapped(a: object, b: object) -> SomeWrap:
#     ...


# def test_wrap_function():
#     result = fn_wrapped(SomeWrap.create(), create_expr())
#     assert result == SomeWrap(fn, (SomeExpr.create(), create_expr()))


# T = typing.TypeVar("T")

# T_expr = typing.TypeVar("T_expr", bound="Expr")


# class Expr(Expression, typing.Generic[T]):
#     @classmethod
#     @expression
#     def create(cls: typing.Type[T_expr]) -> T_expr:
#         ...


# class WrapExpr(Expr[int]):
#     ...


# def test_wrap_returns_generic():
#     assert WrapExpr.create() == WrapExpr(Expr.create, ())


# T_number = typing.TypeVar("T_number", bound="Number")


# class Number(Expression):
#     @expression
#     def __add__(self, other: Number) -> Number:
#         ...

#     @classmethod
#     @expression
#     def create(cls: typing.Type[T_number]) -> T_number:
#         ...


# class WrapNumber(Number):
#     @wrap_method
#     def __add__(self, other: object) -> WrapNumber:
#         ...


# def test_wrap_method():
#     n = WrapNumber.create()
#     k = Number.create()
#     assert n + k == WrapNumber(Number.__add__, (Number.create(), Number.create()))


# class List(Expression, typing.Generic[T]):
#     @expression
#     def __getitem__(self, idx: Number) -> T:
#         ...


# @expression
# def create_list_number() -> List[Number]:
#     ...


# class WrapListInt(Wrap[List[Number]]):
#     @wrap(List.__getitem__)
#     def __getitem__(self, idx: object) -> WrapNumber:
#         ...


# def test_wrap_method_generic():
#     l = create_list_number()
#     i = create_number()
#     assert WrapListInt(l)[i] == WrapNumber(l[i])



import metadsl
import metadsl.python.pure as py_pure
import metadsl.numpy.pure as np_pure


a = np_pure.arange(
    py_pure.Optional.none(py_pure.Number),
    py_pure.Number.from_number(10),
    py_pure.Optional.none(py_pure.Number),
    py_pure.Optional.none(np_pure.DType),
)

a[
    py_pure.Union.left(
        py_pure.Integer, py_pure.Tuple[py_pure.Integer], py_pure.Integer.from_int(10)
    )
]

a[
    py_pure.Union.right(
        py_pure.Integer,
        py_pure.Tuple[py_pure.Integer],
        py_pure.Tuple.from_items(
            py_pure.Integer, py_pure.Integer.from_int(10), py_pure.Integer.from_int(10)
        ),
    )
]


@metadsl.pure_rule
def getitem_condense(a: np_pure.NDArray, left_idx: py_pure.Integer, right_idx: py_pure.Integer):
    return (
        a[
            py_pure.Union.left(
                py_pure.Integer, py_pure.Tuple[py_pure.Integer], left_idx
            )
        ][
            py_pure.Union.left(
                py_pure.Integer, py_pure.Tuple[py_pure.Integer], right_idx
            )
        ],
        a[
            py_pure.Union.right(
                py_pure.Integer,
                py_pure.Tuple[py_pure.Integer],
                py_pure.Tuple.from_items(py_pure.Integer, left_idx, right_idx),
            )
        ],
    )



"""
Compile LLVM module from IR.
"""
from __future__ import annotations
import typing

from metadsl import *
from metadsl_core import *
import llvmlite.binding as binding

__all__ = ["ModuleRef", "ExecutionEngine", "llvmlite_binding_rules"]

binding.initialize()
binding.initialize_native_target()
binding.initialize_native_asmprinter()
target_machine = binding.Target.from_default_triple().create_target_machine()


llvmlite_binding_rules = RulesRepeatFold()
register_llvmlite_binding = llvmlite_binding_rules.append


class ModuleRef(Expression):
    @expression
    @classmethod
    def create(cls, code: str) -> ModuleRef:
        ...

    @expression
    def optimize(self, level: int) -> ModuleRef:
        ...

    @expression
    @classmethod
    def box(cls, mod: binding.ModuleRef) -> ModuleRef:
        ...


@register_llvmlite_binding
@rule
def module_ref_create(code: str) -> R[ModuleRef]:
    def inner() -> ModuleRef:
        llmod = binding.parse_assembly(code)
        llmod.verify()
        return ModuleRef.box(llmod)

    return ModuleRef.create(code), inner


@register_llvmlite_binding
@rule
def module_ref_optimize(mod: binding.ModuleRef, level: int) -> R[ModuleRef]:
    def inner() -> ModuleRef:
        pmb = binding.create_pass_manager_builder()
        pmb.opt_level = level
        pm = binding.create_module_pass_manager()
        pmb.populate(pm)
        pm.run(mod)
        return ModuleRef.box(mod)

    return ModuleRef.box(mod).optimize(level), inner


class ExecutionEngine(Expression):
    @expression
    @classmethod
    def create(cls, mod: ModuleRef) -> ExecutionEngine:
        ...

    @expression
    @classmethod
    def box(cls, engine: binding.ExecutionEngine) -> ExecutionEngine:
        ...

    @expression
    def get_function_address(self, name: str) -> int:
        ...


# keep a global mapping of these so they dont get GCed
_execution_engines: typing.List[binding.ExecutionEngine] = []


@register_llvmlite_binding
@rule
def execution_engine_create(mod: binding.ModuleRef) -> R[ExecutionEngine]:
    def inner() -> ExecutionEngine:
        ee = binding.create_mcjit_compiler(mod, target_machine)
        ee.finalize_object()
        _execution_engines.append(ee)
        return ExecutionEngine.box(ee)

    return ExecutionEngine.create(ModuleRef.box(mod)), inner


@register_llvmlite_binding
@rule
def execution_engine_address(engine: binding.ExecutionEngine, name: str) -> R[int]:
    return (
        ExecutionEngine.box(engine).get_function_address(name),
        lambda: engine.get_function_address(name),
    )



from __future__ import annotations
import typing
import ctypes


from metadsl import *
from metadsl_core import *

__all__ = ["ctypes_rules", "CType", "CFunctionType"]


ctypes_rules = RulesRepeatFold()
register_ctypes = ctypes_rules.append


class CType(Expression):
    @expression
    @classmethod
    def c_int(cls) -> CType:
        ...

    @expression
    @classmethod
    def box(cls, tp: typing.Any) -> CType:
        ...


@register_ctypes
@rule
def ctype_cint() -> R[CType]:
    return CType.c_int(), lambda: CType.box(ctypes.c_int)


class CFunctionType(Expression):
    @expression
    @classmethod
    def create(cls, return_type: CType, *args: CType) -> CFunctionType:
        ...

    # Cannot represent type of C function
    @expression
    @classmethod
    def box(cls, tp: typing.Any) -> CFunctionType:
        ...

    @expression
    def __call__(self, ptr: int) -> typing.Callable:
        ...


@register_ctypes
@rule
def c_function_type_create_1(
    return_tp: typing.Any, arg1: typing.Any
) -> R[CFunctionType]:
    return (
        CFunctionType.create(CType.box(return_tp), CType.box(arg1)),
        lambda: CFunctionType.box(ctypes.CFUNCTYPE(return_tp, arg1)),
    )


@register_ctypes
@rule
def c_function_type_create_2(
    return_tp: typing.Any, arg1: typing.Any, arg2: typing.Any
) -> R[CFunctionType]:
    return (
        CFunctionType.create(CType.box(return_tp), CType.box(arg1), CType.box(arg2)),
        lambda: CFunctionType.box(ctypes.CFUNCTYPE(return_tp, arg1, arg2)),
    )


@register_ctypes
@rule
def cfunc_call(fntype: typing.Any, ptr: int) -> R[typing.Callable]:
    return CFunctionType.box(fntype)(ptr), lambda: fntype(ptr)



import typing
from metadsl_core import rule_groups

from .ctypes import ctypes_rules
from .integration import llvm_integration_rules
from .ir_llvmlite import ir_llvmlite_rules
from .llvmlite_binding import llvmlite_binding_rules
from .ir_context import ir_context_rules

__all__: typing.List[str] = []


# First do higher level context rules
rule_groups["llvmlite.context"] = ir_context_rules
# First do integration rules so that all functions are created
rule_groups["llvmlite.integration"] = llvm_integration_rules
# Then create the llvm modules
rule_groups["llvmlite.ir_llvmlite"] = ir_llvmlite_rules
# Finally you can create the ctypes
rule_groups["llvmlite.ctypes"] = ctypes_rules
# And then compile into a function
rule_groups["llvmlite.llvmlite_binding"] = llvmlite_binding_rules



from __future__ import annotations
import typing

from metadsl import *
from .abstraction import *
from .rules import *
from .pair import *

__all__ = ["Maybe"]
T = typing.TypeVar("T")
U = typing.TypeVar("U")


class Maybe(Expression, typing.Generic[T]):
    """
    An optional value.
    """

    @expression
    @classmethod
    def nothing(cls) -> Maybe[T]:
        ...

    @expression
    @classmethod
    def just(cls, value: T) -> Maybe[T]:
        ...

    @expression
    def match(self, nothing: U, just: Abstraction[T, U]) -> U:
        ...

    @expression
    def __or__(self, other: Maybe[T]) -> Maybe[T]:
        """
        Like the <|> function https://en.wikibooks.org/wiki/Haskell/Alternative_and_MonadPlus
        """
        ...

    @expression
    def __and__(self, other: Maybe[U]) -> Maybe[Pair[T, U]]:
        """
        >>> execute(Maybe[int].nothing() & Maybe.just("")) == Maybe[Pair[int, str]].nothing()
        True
        >>> execute(Maybe.just(10) & Maybe[str].nothing()) == Maybe[Pair[int, str]].nothing()
        True
        >>> execute(Maybe.just(10) & Maybe.just("")) == Maybe.just(Pair.create(10, ""))
        True
        """
        ...

    @expression
    def map(self, just: Abstraction[T, U]) -> Maybe[U]:
        return self.match(
            Maybe[U].nothing(),
            Abstraction.from_fn(Maybe[U].just) + just,  # type: ignore
        )


register(default_rule(Maybe[T].map))


@register  # type: ignore
@rule
def maybe_match(nothing: U, just: Abstraction[T, U], v: T) -> R[U]:
    yield Maybe[T].nothing().match(nothing, just), nothing
    yield Maybe.just(v).match(nothing, just), just(v)


@register
@rule
def maybe_or(x: T, v: Maybe[T]) -> R[Maybe[T]]:
    yield Maybe[T].nothing() | Maybe[T].nothing(), Maybe[T].nothing()
    yield Maybe.just(x) | v, Maybe.just(x)
    yield v | Maybe.just(x), Maybe.just(x)


@register
@rule
def maybe_and(left: Maybe[T], right: Maybe[U], left_v: T, right_v: U) -> R[Pair[T, U]]:
    yield Maybe[T].nothing() & right, Maybe[Pair[T, U]].nothing()
    yield left & Maybe[U].nothing(), Maybe[Pair[T, U]].nothing()
    yield Maybe.just(left_v) & Maybe.just(right_v), Maybe.just(
        Pair.create(left_v, right_v)
    )



from __future__ import annotations
import typing
import pytest

from .matching import *
from .rules import *
from .typing_tools import *
from .expressions import *


class _SomeExpression(Expression):
    pass


class TestWildcard:
    def test_create_literal(self):
        w = create_wildcard(int)  # type: ignore
        assert get_type(w) == PlaceholderExpression[int]  # type: ignore

    def test_create_expression(self):
        w = create_wildcard(_SomeExpression)
        assert get_type(w) == _SomeExpression


class _Number(Expression):
    @expression
    def __add__(self, other: _Number) -> _Number:
        ...

    @expression
    @classmethod
    def NaN(cls) -> _Number:
        ...


@expression
def _from_int(i: int) -> _Number:
    ...


@rule
def _add_rule(a: int, b: int) -> R[_Number]:
    return (_from_int(a) + _from_int(b), lambda: _from_int(a + b))


T = typing.TypeVar("T")


class _List(Expression, typing.Generic[T]):
    @expression
    def __add__(self, other: _List[T]) -> _List[T]:
        ...

    @expression
    @classmethod
    def create(cls, *items: T) -> _List[T]:
        ...

    @staticmethod
    @expression
    def sum(l: _List[_Number]) -> _Number:
        ...


U = typing.TypeVar("U")


class Both(Expression, typing.Generic[T, U]):
    @expression
    @classmethod
    def create(cls) -> Both[T, U]:
        ...


V = typing.TypeVar("V")


class Abstraction(Expression, typing.Generic[T, U]):
    @expression
    @classmethod
    def create(cls) -> Abstraction[T, U]:
        ...

    @expression
    def __add__(self, other: Abstraction[U, V]) -> Abstraction[T, V]:
        ...


class TestRule:
    def test_add(self):
        expr = _from_int(1) + _from_int(2)
        assert execute(expr, _add_rule) == _from_int(3)

    def test_type_args(self):
        @rule
        def _concat_lists(l: T, r: T) -> R[_List[T]]:
            return _List.create(l) + _List.create(r), lambda: _List.create(l, r)

        assert execute(
            _List.create(1) + _List.create(2), _concat_lists
        ) == _List.create(1, 2)

    def test_variable_args(self):
        @rule
        def _concat_lists(
            ls: typing.Sequence[T], rs: typing.Sequence[T]
        ) -> R[_List[T]]:
            return (
                _List[T].create(*ls) + _List[T].create(*rs),
                lambda: _List[T].create(*ls, *rs),
            )

        assert execute(_List.create(1, 2) + _List.create(3, 4), _concat_lists) == _List[
            int
        ].create(1, 2, 3, 4)

    def test_variable_args_empty(self):
        @rule
        def _concat_lists(
            ls: typing.Sequence[T], rs: typing.Sequence[T]
        ) -> R[_List[T]]:
            return (
                _List[T].create(*ls) + _List[T].create(*rs),
                lambda: _List[T].create(*ls, *rs),
            )

        assert (
            execute(_List[int].create() + _List[int].create(), _concat_lists)
            == _List[int].create()
        )

    def test_variable_args_right_side(self):
        @rule
        def _concat_lists_minus_end(
            ls: typing.Sequence[T], rs: typing.Sequence[T], l: T, r: T
        ) -> R[_List[T]]:
            return (
                _List[T].create(*ls, l) + _List[T].create(*rs, r),
                lambda: _List[T].create(*ls, *rs),
            )

        assert execute(
            _List.create(1, 2) + _List.create(3, 4), _concat_lists_minus_end
        ) == _List[int].create(1, 3)

        assert (
            execute(_List.create(1) + _List.create(3), _concat_lists_minus_end)
            == _List[int].create()
        )

    def test_variable_args_two_left(self):
        @rule
        def _remove_one_first(xs: typing.Sequence[int], x: int) -> R[_List[int]]:
            return (
                _List.create(1, x, *xs),
                lambda: _List.create(x, *xs),
            )

        assert execute(_List.create(1, 2, 2), _remove_one_first) == _List.create(2, 2)
        assert execute(_List.create(1, 2), _remove_one_first) == _List.create(2)
        assert execute(_List.create(1), _remove_one_first) == _List.create(1)

    @pytest.mark.skip("This isn't supported yet")
    def test_variable_args_generator(self):
        """
        We should be able to iterate through variables args to create a sequence
        of homogenous rules
        """

        @rule
        def _sum_list(xs: typing.Sequence[int]) -> R[_Number]:
            return (
                _List.sum(_List[_Number].create(*(_from_int(x) for x in xs))),
                lambda: _from_int(sum(xs)),
            )

        assert execute(_List.sum(_List[_Number].create()), _sum_list) == _from_int(0)
        assert execute(_List.sum(_List.create(_from_int(10))), _sum_list) == _from_int(
            10
        )
        assert execute(
            _List.sum(_List.create(_from_int(10), _from_int(1))), _sum_list
        ) == _from_int(1)

        # Verify doesn't replace when not integers
        assert execute(_List.sum(_List.create(_Number.NaN())), _sum_list) == _List.sum(
            _List.create(_Number.NaN())
        )
        assert execute(
            _List.sum(_List.create(_from_int(0), _Number.NaN())), _sum_list
        ) == _List.sum(_List.create(_from_int(0), _Number.NaN()))

        assert execute(
            _List.sum(_List.create(_Number.NaN(), _from_int(0))), _sum_list
        ) == _List.sum(_List.create(_Number.NaN(), _from_int(0)))

    def test_different_generic_param(self):
        """
        For this test, we want to make sure if we use the saem TypeVar instance in our match rule
        that we did in our expresion, it can be seperated.
        """
        # Here we have `_List[T]` work for the type var `T` in the list
        @rule
        def _add_list_of_lists(
            ls: typing.Sequence[T], rs: typing.Sequence[T]
        ) -> R[_List[_List[T]]]:
            return (
                _List.create(_List[T].create(*ls)) + _List.create(_List[T].create(*rs)),
                lambda: _List.create(_List[T].create(*ls, *rs)),
            )

        assert execute(
            _List.create(_List.create(1, 2)) + _List.create(_List.create(3, 4)),
            _add_list_of_lists,
        ) == _List.create(_List[int].create(1, 2, 3, 4))
        assert execute(
            _List.create(_List[int].create()) + _List.create(_List[int].create()),
            _add_list_of_lists,
        ) == _List.create(_List[int].create())

    def test_non_lambda_result(self):
        @expression
        def _from_str(s: str) -> _Number:
            ...

        @rule
        def _add_zero_rule(a: _Number) -> R[_Number]:
            return a + _from_int(0), a

        s = _from_str("str")
        expr = s + _from_int(0)
        assert execute(expr, _add_zero_rule) == s

    def test_generator_rule(self):
        @expression
        def _from_str(s: str) -> _Number:
            ...

        @rule
        def _add_zero_rule(a: _Number) -> R[_Number]:
            yield a + _from_int(0), a
            yield _from_int(0) + a, a

        s = _from_str("str")
        assert execute(s + _from_int(0), _add_zero_rule) == s
        assert execute(_from_int(0) + s, _add_zero_rule) == s

    def test_generator_rule_different_wildcards(self):
        @expression
        def _from_str(s: str) -> _Number:
            ...

        @rule
        def _add_zero_rule(a: _Number, b: _Number) -> R[_Number]:
            yield a + _from_int(0), a
            yield _from_int(0) + b, b

        s = _from_str("str")
        assert execute(s + _from_int(0), _add_zero_rule) == s
        assert execute(_from_int(0) + s, _add_zero_rule) == s

    def test_two_params(self):
        """
        For this test, we want to make sure if we use the saem TypeVar instance in our match rule
        that we did in our expresion, it can be seperated.
        """

        @expression
        def combine_lists(l: _List[T], r: _List[U]) -> _List[Both[T, U]]:
            ...

        # Here we have `_List[T]` work for the type var `T` in the list
        @rule
        def combine_lists_rule() -> R[_List[Both[T, U]]]:
            return (
                combine_lists(_List[T].create(), _List[U].create()),
                _List[Both[T, U]].create(),
            )

        assert (
            execute(
                combine_lists(_List[int].create(), _List[float].create()),
                combine_lists_rule,
            )
            == _List[Both[int, float]].create()
        )

    def test_two_different_params(self):
        """
        For this test, we want to make sure if we use the saem TypeVar instance in our match rule
        that we did in our expresion, it can be seperated.
        """

        # Here we have `_List[T]` work for the type var `T` in the list
        @rule
        def add_abstractions_rule() -> R[Abstraction[T, V]]:
            return (  # type: ignore
                Abstraction[T, U].create() + Abstraction[U, V].create(),
                Abstraction[T, V].create(),
            )

        assert (
            execute(
                Abstraction[int, float].create() + Abstraction[float, str].create(),
                add_abstractions_rule,
            )
            == Abstraction[int, str].create()
        )

    def test_literal_arg(self):
        """
        If we pass in a non expression arg, the rule shouldn't execute until it is really an instance
        of that, not an instance of expression.
        """

        @expression
        def add(i: int, j: int) -> int:
            ...

        @expression
        def subtract(i: int, j: int) -> int:
            ...

        @rule  # type: ignore
        def add_rule(i: int, j: int) -> R[int]:
            return add(i, j), lambda: i + j

        @rule  # type: ignore
        def subtract_rule(i: int, j: int) -> R[int]:
            return subtract(i, j), lambda: i - j

        assert execute(subtract(add(1, 2), 3), subtract_rule) == subtract(add(1, 2), 3)

        assert execute(subtract(add(1, 2), 3), RulesRepeatFold(add_rule)) == subtract(
            3, 3
        )
        assert (
            execute(subtract(add(1, 2), 3), RulesRepeatFold(add_rule, subtract_rule))
            == 0
        )


class TestDefaultRule:
    def test_fn(self):
        @expression
        def inner_fn(a: int) -> Expression:
            ...

        @expression
        def fn(a: int) -> Expression:
            return inner_fn(a)

        assert fn(10) != inner_fn(10)
        assert execute(fn(10), default_rule(fn)) == inner_fn(10)

    def test_method(self):
        class C(Expression):
            @expression
            def __add__(self, other: C) -> C:
                ...

            @expression
            def double(self) -> C:
                return self + self

        globals()["C"] = C

        @expression
        def create() -> C:
            ...

        rule = default_rule(C.double)
        assert create().double() != create() + create()
        assert execute(create().double(), rule) == create() + create()

    def test_method_generic(self):
        class C(Expression, typing.Generic[T]):
            @expression
            def __add__(self, other: C[T]) -> C[T]:
                ...

            @expression
            def double(self) -> C[T]:
                return self + self

            @expression
            @classmethod
            def create(cls) -> C[T]:
                ...

        globals()["C"] = C

        expr = C[int].create().double()

        rule = default_rule(C.double)
        assert execute(expr, rule) == C[int].create() + C[int].create()

        assert execute(expr, rule) != C[str].create() + C[str].create()
        assert expr != C[int].create() + C[int].create()

    def test_classmethod_generic(self):
        class C(Expression, typing.Generic[T]):
            @expression
            @classmethod
            def create(cls) -> C[T]:
                ...

            @expression
            @classmethod
            def create_wrapper(cls) -> C[T]:
                return cls.create()

        globals()["C"] = C

        expr = C[int].create_wrapper()
        rule = default_rule(C[T].create_wrapper)

        assert execute(expr, rule) == C[int].create()

        assert execute(expr, rule) != C[str].create()
        assert expr != C[int].create()

    def test_classmethod_generic_arg(self):
        class C(Expression, typing.Generic[T]):
            @expression
            @classmethod
            def create(cls, v: T) -> C[T]:
                ...

            @expression
            @classmethod
            def create_wrapper(cls, v: T) -> C[T]:
                return cls.create(v)

        globals()["C"] = C

        expr = C.create_wrapper(123)
        rule = default_rule(C.create_wrapper)
        assert execute(expr, rule) == C.create(123)

        assert expr != C.create(123)

    def test_classmethod_generic_arg_different_var(self):
        class C(Expression, typing.Generic[T]):
            @expression
            @classmethod
            def create(cls) -> C[T]:
                ...

            @expression
            @classmethod
            def create_wrapper(cls, v: V) -> C[V]:
                return C[V].create()

        globals()["C"] = C

        expr = C.create_wrapper(123)
        rule = default_rule(C.create_wrapper)
        assert execute(expr, rule) == C[int].create()

        assert expr != C[int].create()

    def test_function_generic(self):
        @expression
        def identity(a: T) -> T:
            return a

        rule = default_rule(identity)
        assert execute(identity(1), rule) == 1
        assert execute(identity(_from_int(1)), rule) == _from_int(1)

    def test_function_generic_call_generic(self):
        @expression
        def identity(a: T) -> T:
            ...

        @expression
        def identity_2(a: T) -> T:
            return identity(a)

        rule = default_rule(identity_2)
        assert execute(identity_2(1), rule) == identity(1)
        assert execute(identity_2(_from_int(1)), rule) == identity(_from_int(1))

    def test_literal_arg(self):
        """
        If we pass in a non expression arg, the rule shouldn't execute until it is really an instance
        of that, not an instance of expression.
        """

        @expression
        def add(i: int, j: int) -> int:
            return i + j

        @expression
        def subtract(i: int, j: int) -> int:
            return i - j

        add_rule = default_rule(add)
        subtract_rule = default_rule(subtract)

        assert execute(subtract(add(1, 2), 3), subtract_rule) == subtract(add(1, 2), 3)

        assert execute(subtract(add(1, 2), 3), RulesRepeatFold(add_rule)) == subtract(
            3, 3
        )
        assert (
            execute(subtract(add(1, 2), 3), RulesRepeatFold(subtract_rule, add_rule))
            == 0
        )

    def test_property(self):
        class C(Expression, typing.Generic[T]):
            @expression
            @classmethod
            def create(cls) -> C[T]:
                ...

            @property  # type: ignore
            @expression
            def get(self) -> T:
                ...

        globals()["C"] = C



"""
Generate LLVM IR
"""

from __future__ import annotations
import typing

import llvmlite.ir as ir

from metadsl import *
from metadsl_core import *


__all__ = [
    "ModRef",
    "FnRef",
    "Type",
    "FnType",
    "BlockRef",
    "Value",
    "Terminate",
    "Fn",
    "Mod",
    "_Uniq",
]


class Mod(Expression):
    @expression  # type: ignore
    @property
    def ref(self) -> ModRef:
        ...

    @expression
    def add_fn(self, fn: Fn) -> Mod:
        ...


class Fn(Expression):
    @expression
    def add_block(self, terminate: Terminate) -> Fn:
        ...

    @expression  # type: ignore
    @property
    def ref(self) -> FnRef:
        ...


class Terminate(Expression):
    """
    Returned at the end of a block.
    """

    pass


class Value(Expression):
    @expression
    @classmethod
    def constant(cls, type: Type, value: typing.Any) -> Value:
        ...


class BlockRef(Expression):
    @expression
    def ret(self, value: Value) -> Terminate:
        ...

    @expression
    def icmp_signed(self, operator: str, left: Value, right: Value) -> Value:
        ...

    @expression
    def call(self, function: FnRef, args: Vec[Value]) -> Value:
        ...

    @expression
    def cbranch(self, condition: Value, true: BlockRef, false: BlockRef) -> Terminate:
        ...

    @expression
    def branch(self, target: BlockRef) -> Terminate:
        ...

    @expression
    def add(self, left: Value, right: Value) -> Value:
        ...

    @expression
    def sub(self, left: Value, right: Value) -> Value:
        ...

    @expression
    def phi(self, *incoming: Pair[Value, BlockRef]) -> Value:
        ...


class _Uniq:
    """
    Unique value that should only be equal to itself.

    Used so that multiple blocks can each be created with a different
    object so they will not be combined.
    """

    def __eq__(self, value) -> bool:
        return self is value

    def __str__(self) -> str:
        return str(id(self))

    def __repr__(self) -> str:
        return f"#{self}"


class FnRef(Expression):
    @expression
    def fn(self, *blocks: Terminate) -> Fn:
        ...

    def block(self, is_first: bool, name: typing.Union[str, None] = None) -> BlockRef:
        return self._block(is_first, name, _Uniq())

    @expression
    def _block(
        self, is_first: bool, name: typing.Union[str, None], uniq: _Uniq
    ) -> BlockRef:
        ...

    @expression  # type: ignore
    @property
    def arguments(self) -> Vec[Value]:
        ...

    @expression  # type: ignore
    @property
    def name(self) -> str:
        ...

    @expression  # type: ignore
    @property
    def type(self) -> FnType:
        ...


class ModRef(Expression):
    @expression
    @classmethod
    def create(cls, name: str) -> ModRef:
        ...

    @expression
    def fn(self, name: str, type: FnType, calling_convention: str = "") -> FnRef:
        ...

    @expression
    def mod(self, *fns: Fn) -> Mod:
        ...


class FnType(Expression):
    @expression
    @classmethod
    def create(cls, return_type: Type, *arg_types: Type) -> FnType:
        ...


class Type(Expression):
    @expression
    @classmethod
    def create_int(cls, bits: int) -> Type:
        ...


@register
@rule
def mod_add_fn(ref: ModRef, fns: typing.Sequence[Fn], fn: Fn):
    return (ref.mod(*fns).add_fn(fn), ref.mod(*fns, fn))


@register
@rule
def mod_ref(ref: ModRef, fns: typing.Sequence[Fn]):
    return (ref.mod(*fns).ref, ref)


@register
@rule
def fn_add_block(ref: FnRef, blocks: typing.Sequence[Terminate], block: Terminate):
    return (ref.fn(*blocks).add_block(block), ref.fn(*blocks, block))


@register
@rule
def fn_ref(ref: FnRef, blocks: typing.Sequence[Terminate]):
    return (ref.fn(*blocks).ref, ref)


@register
@rule
def fn_ref_get_name(mod: ModRef, name: str, tp: FnType, calling_convention: str):
    return mod.fn(name, tp, calling_convention).name, name


@register
@rule
def fn_ref_get_type(mod: ModRef, name: str, tp: FnType, calling_convention: str):
    return mod.fn(name, tp, calling_convention).type, tp



from __future__ import annotations

import typing

from metadsl import *
from .rules import *
from .abstraction import *
from .abstraction import Variable, from_fn_rule
from .integer import *


T = typing.TypeVar("T")


class I(Expression):
    @expression
    @classmethod
    def create(cls, i: int) -> I:
        ...

    @expression
    def inc(self) -> I:
        ...

    @expression
    def dec(self) -> I:
        ...


class TestAbstraction:
    def test_from_fn(self):
        a = execute(Abstraction[typing.Any, int].from_fn(lambda i: i), from_fn_rule)
        assert isinstance(a, Abstraction)
        assert a.function == Abstraction[typing.Any, int].create
        var, body = a.args
        assert (
            var.function == Abstraction[typing.Any, int].create_variable  # type: ignore
        )
        (var_,) = var.args  # type: ignore
        assert isinstance(var_, Variable)
        assert body == var

    def test_constant(self):
        assert execute(Abstraction[typing.Any, int].from_fn(lambda _: 10)("hi")) == 10

    def test_identity(self):
        assert execute(Abstraction[str, str].from_fn(lambda i: i)("hi")) == "hi"

    def test_compose(self):
        assert execute(
            (Abstraction.from_fn(I.inc) + Abstraction.from_fn(I.dec))(I.create(10))
        ) == I.inc(I.dec(I.create(10)))

    def test_fix(self):
        one = Integer.from_int(1)
        zero = Integer.from_int(0)

        @Abstraction.fix
        @Abstraction.from_fn
        def factorial(
            fact_fn: Abstraction[Integer, Integer]
        ) -> Abstraction[Integer, Integer]:
            @Abstraction.from_fn
            def inner(n: Integer) -> Integer:
                return n.eq(zero).if_(one, n * fact_fn(n - one))

            return inner

        assert execute(factorial(zero)) == one
        assert execute(factorial(one)) == one
        assert execute(factorial(Integer.from_int(2))) == Integer.from_int(2)
        assert execute(factorial(Integer.from_int(3))) == Integer.from_int(6)

    def test_unfix_fixed(self):
        one = Integer.from_int(1)
        zero = Integer.from_int(0)

        @Abstraction.fix
        @Abstraction.unfix
        @Abstraction.fix
        @Abstraction.from_fn
        def factorial(
            fact_fn: Abstraction[Integer, Integer]
        ) -> Abstraction[Integer, Integer]:
            @Abstraction.from_fn
            def inner(n: Integer) -> Integer:
                return n.eq(zero).if_(one, n * fact_fn(n - one))

            return inner

        assert execute(factorial(zero)) == one
        assert execute(factorial(one)) == one

    def test_unfix_not_fixed(self):
        two = Integer.from_int(2)
        one = Integer.from_int(1)
        zero = Integer.from_int(0)

        @Abstraction.from_fn
        def add_one(i: Integer) -> Integer:
            return i + one

        @Abstraction.from_fn
        def add_two(i: Integer) -> Integer:
            return i + one + one

        assert execute(Abstraction.unfix(add_one)(add_one)(zero)) == one
        assert execute(Abstraction.unfix(add_two)(add_one)(zero)) == two



"""
LLVM support.
"""

__version__ = "0.2.0"


from .ctypes import *
from .integration import *
from .ir_llvmlite import *
from .ir import *
from .llvmlite_binding import *
from .ir_context import *
from . import rules as __llvmlite_rules  # NOQA



import collections.abc
import dataclasses
import typing

__all__ = ["safe_merge", "Item", "UnhashableMapping", "HashableMapping"]

T = typing.TypeVar("T")
V = typing.TypeVar("V")


@dataclasses.dataclass
class Item(typing.Generic[T, V]):
    key: T
    value: V


@dataclasses.dataclass(init=False, frozen=True)
class HashableMapping(collections.abc.Mapping, typing.Generic[T, V]):
    """
    Like a dictionary, but immutable and hashable.
    """

    _items: typing.Tuple[typing.Tuple[T, V], ...]

    def __init__(self, mapping: typing.Mapping[T, V]):
        object.__setattr__(
            self, "_items", tuple((k, v) for k, v in mapping.items())
        )

    def __hash__(self):
        return hash(self._items)

    def __getitem__(self, key: T) -> V:
        for item in self._items:
            if item[0] == key:
                return item[1]
        raise KeyError()

    def __iter__(self):
        for item in self._items:
            yield item[0]

    def __len__(self):
        return len(self._items)


@dataclasses.dataclass(init=False)
class UnhashableMapping(collections.abc.MutableMapping, typing.Generic[T, V]):
    """
    Like a dictionary, but can have unhashable keys.
    """

    _items: typing.List[Item[T, V]]

    def __init__(self, *items: Item[T, V]):
        self._items = list(items)

    def __getitem__(self, key: T) -> V:
        item = self._find_item(key)
        return item.value

    def __setitem__(self, key: T, value: V) -> None:
        try:
            item = self._find_item(key)
        except KeyError:
            item = Item(key, value)
            self._items.append(item)
        else:
            item.value = value

    def __delitem__(self, key: T) -> None:
        self._items.remove(self._find_item(key))

    def _find_item(self, key: T) -> Item[T, V]:
        for item in self._items:
            if item.key == key:
                return item
        raise KeyError()

    def __iter__(self):
        for item in self._items:
            yield item.key

    def __len__(self):
        return len(self._items)


def safe_merge(
    *mappings: typing.Mapping[T, V],
    dict_constructor: typing.Type[typing.MutableMapping] = dict,
) -> typing.Mapping[T, V]:
    """
    Combing mappings by merging the dictionaries. It raises a ValueError
    if there are duplicate keys and they are not equal.
    """
    res: typing.MutableMapping[T, V] = dict_constructor()
    for mapping in mappings:
        for key, value in mapping.items():
            if key in res:
                if res[key] != value:
                    raise ValueError()
            else:
                res[key] = value
    return res
